# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""A task for sending EV Charger power bounds to the microgrid API."""

import asyncio
import logging
from dataclasses import dataclass
from datetime import timedelta

from frequenz.channels import Broadcast, LatestValueCache, Sender, select, selected_from
from frequenz.channels.timer import SkipMissedAndDrift, Timer
from frequenz.client.microgrid import ComponentCategory, MeterData

from ..._internal._asyncio import cancel_and_await
from ...microgrid import connection_manager

_logger = logging.getLogger(__name__)


@dataclass
class ComponentCurrentLimit:
    """A current limit, to be sent to the EV Charger."""

    component_id: int
    """The component ID of the EV Charger."""

    max_amps: float
    """The maximum current in amps, that an EV can draw from this EV Charger."""


class BoundsSetter:
    """A task for sending EV Charger power bounds to the microgrid API.

    Also, periodically resends the last set bounds to the microgrid API, if no new
    bounds have been set.
    """

    _NUM_PHASES = 3
    """Number of phases in the microgrid."""

    def __init__(self, repeat_interval: timedelta) -> None:
        """Create a `BoundsSetter` instance.

        Args:
            repeat_interval: Interval after which to repeat the last set bounds to the
                microgrid API, if no new calls to `set_bounds` have been made.
        """
        self._repeat_interval = repeat_interval

        self._task: asyncio.Task[None] = asyncio.create_task(self._run())
        self._bounds_chan: Broadcast[ComponentCurrentLimit] = Broadcast(
            name="BoundsSetter"
        )
        self._bounds_rx = self._bounds_chan.new_receiver()
        self._bounds_tx = self._bounds_chan.new_sender()
        self._meter_data_cache: LatestValueCache[MeterData] | None = None

    async def set(self, component_id: int, max_amps: float) -> None:
        """Send the given current limit to the microgrid for the given component id.

        Args:
            component_id: ID of EV Charger to set the current bounds to.
            max_amps: maximum current in amps, that an EV can draw from this EV Charger.
        """
        await self._bounds_tx.send(ComponentCurrentLimit(component_id, max_amps))

    def new_bounds_sender(self) -> Sender[ComponentCurrentLimit]:
        """Return a `Sender` for setting EV Charger current bounds with.

        Returns:
            A new `Sender`.
        """
        return self._bounds_chan.new_sender()

    async def stop(self) -> None:
        """Stop the BoundsSetter."""
        if self._meter_data_cache is not None:
            await self._meter_data_cache.stop()
        await self._bounds_chan.close()
        await cancel_and_await(self._task)

    async def _run(self) -> None:
        """Wait for new bounds and forward them to the microgrid API.

        Also, periodically resend the last set bounds to the microgrid API, if no new
        bounds have been set.

        Raises:
            RuntimeError: If no meters are found in the component graph.
            ValueError: If the meter channel is closed.
        """
        api_client = connection_manager.get().api_client
        graph = connection_manager.get().component_graph
        meters = graph.components(component_categories={ComponentCategory.METER})
        if not meters:
            err = "No meters found in the component graph."
            _logger.error(err)
            raise RuntimeError(err)

        meter_id = next(iter(meters)).component_id
        self._meter_data_cache = LatestValueCache(
            await api_client.meter_data(meter_id),
            unique_id=f"{type(self).__name__}«{hex(id(self))}»:meter«{meter_id}»",
        )
        latest_bound: dict[int, ComponentCurrentLimit] = {}

        bound_chan = self._bounds_rx
        timer = Timer(
            timedelta(self._repeat_interval.total_seconds()), SkipMissedAndDrift()
        )

        async for selected in select(bound_chan, timer):
            meter = self._meter_data_cache.get()
            if meter is None:
                raise ValueError("Meter channel closed.")

            if selected_from(selected, bound_chan):
                bound: ComponentCurrentLimit = selected.message
                if (
                    bound.component_id in latest_bound
                    and latest_bound[bound.component_id] == bound
                ):
                    continue
                latest_bound[bound.component_id] = bound
                min_voltage = min(meter.voltage_per_phase)
                _logger.info("sending new bounds: %s", bound)
                await api_client.set_bounds(
                    bound.component_id,
                    0,
                    bound.max_amps * min_voltage * self._NUM_PHASES,
                )
            elif selected_from(selected, timer):
                for bound in latest_bound.values():
                    min_voltage = min(meter.voltage_per_phase)
                    _logger.debug("resending bounds: %s", bound)
                    await api_client.set_bounds(
                        bound.component_id,
                        0,
                        bound.max_amps * min_voltage * self._NUM_PHASES,
                    )



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A logical meter for calculating high level metrics for a microgrid."""

from ._logical_meter import LogicalMeter

__all__ = ["LogicalMeter"]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Read and update logging severity from config."""

import logging
from collections.abc import Mapping
from dataclasses import field
from typing import Annotated, Any, Self, cast

import marshmallow
import marshmallow.validate
from frequenz.channels import Receiver
from marshmallow import RAISE
from marshmallow_dataclass import class_schema, dataclass

from frequenz.sdk.actor import Actor

_logger = logging.getLogger(__name__)

LogLevel = Annotated[
    str,
    marshmallow.fields.String(
        validate=marshmallow.validate.OneOf(choices=logging.getLevelNamesMapping())
    ),
]


@dataclass
class LoggerConfig:
    """A configuration for a logger."""

    level: LogLevel = field(
        default="NOTSET",
        metadata={
            "metadata": {
                "description": "Log level for the logger. Uses standard logging levels."
            },
            "required": False,
        },
    )
    """The log level for the logger."""


@dataclass
class LoggingConfig:
    """A configuration for the logging system."""

    root_logger: LoggerConfig = field(
        default_factory=LoggerConfig,
        metadata={
            "metadata": {
                "description": "Default default configuration for all loggers.",
            },
            "required": False,
        },
    )
    """The default log level."""

    loggers: dict[str, LoggerConfig] = field(
        default_factory=dict,
        metadata={
            "metadata": {
                "description": "Configuration for a logger (the key is the logger name)."
            },
            "required": False,
        },
    )
    """The list of loggers configurations."""

    @classmethod
    def load(cls, configs: Mapping[str, Any]) -> Self:  # noqa: DOC502
        """Load and validate configs from a dictionary.

        Args:
            configs: The configuration to validate.

        Returns:
            The configuration if they are valid.

        Raises:
            ValidationError: if the configuration are invalid.
        """
        schema = class_schema(cls)()
        return cast(Self, schema.load(configs, unknown=RAISE))


class LoggingConfigUpdater(Actor):
    """Actor that listens for logging configuration changes and sets them.

    Example:
        `config.toml` file:
        ```toml
        [logging.root_logger]
        level = "INFO"

        [logging.loggers."frequenz.sdk.actor.power_distributing"]
        level = "DEBUG"

        [logging.loggers."frequenz.channels"]
        level = "DEBUG"
        ```

        ```python
        import asyncio
        from collections.abc import Mapping
        from typing import Any

        from frequenz.channels import Broadcast
        from frequenz.sdk.config import LoggingConfigUpdater, ConfigManager
        from frequenz.sdk.actor import run as run_actors

        async def run() -> None:
            config_channel = Broadcast[Mapping[str, Any]](name="config", resend_latest=True)
            actors = [
                ConfigManager(config_paths=["config.toml"], output=config_channel.new_sender()),
                LoggingConfigUpdater(
                    config_recv=config_channel.new_receiver(limit=1)).map(
                        lambda app_config: app_config.get("logging", {}
                    )
                ),
            ]
            await run_actors(*actors)

        asyncio.run(run())
        ```

        Now whenever the `config.toml` file is updated, the logging configuration
        will be updated as well.
    """

    def __init__(
        self,
        config_recv: Receiver[Mapping[str, Any]],
        log_format: str = "%(asctime)s %(levelname)-8s %(name)s:%(lineno)s: %(message)s",
        log_datefmt: str = "%Y-%m-%dT%H:%M:%S%z",
    ):
        """Initialize this instance.

        Args:
            config_recv: The receiver to listen for configuration changes.
            log_format: Use the specified format string in logs.
            log_datefmt: Use the specified date/time format in logs.

        Note:
            The `log_format` and `log_datefmt` parameters are used in a call to
            `logging.basicConfig()`. If logging has already been configured elsewhere
            in the application (through a previous `basicConfig()` call), then the format
            settings specified here will be ignored.
        """
        super().__init__()
        self._config_recv = config_recv

        # Setup default configuration.
        # This ensures logging is configured even if actor fails to start or
        # if the configuration cannot be loaded.
        self._current_config: LoggingConfig = LoggingConfig()

        logging.basicConfig(
            format=log_format,
            datefmt=log_datefmt,
        )
        self._update_logging(self._current_config)

    async def _run(self) -> None:
        """Listen for configuration changes and update logging."""
        async for message in self._config_recv:
            try:
                new_config = LoggingConfig.load(message)
            except marshmallow.ValidationError:
                _logger.exception(
                    "Invalid logging configuration received. Skipping config update"
                )
                continue

            if new_config != self._current_config:
                self._update_logging(new_config)

    def _update_logging(self, config: LoggingConfig) -> None:
        """Configure the logging level."""
        # If the logger is not in the new config, set it to NOTSET
        loggers_to_unset = self._current_config.loggers.keys() - config.loggers.keys()
        for logger_id in loggers_to_unset:
            _logger.debug("Unsetting log level for logger '%s'", logger_id)
            logging.getLogger(logger_id).setLevel(logging.NOTSET)

        self._current_config = config
        _logger.debug(
            "Setting root logger level to '%s'", self._current_config.root_logger.level
        )
        logging.getLogger().setLevel(self._current_config.root_logger.level)

        # For each logger in the new config, set the log level
        for logger_id, logger_config in self._current_config.loggers.items():
            _logger.debug(
                "Setting log level for logger '%s' to '%s'",
                logger_id,
                logger_config.level,
            )
            logging.getLogger(logger_id).setLevel(logger_config.level)

        _logger.info("Logging config changed to: %s", self._current_config)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""User interface for requesting aggregated battery-inverter data."""


import asyncio
import uuid
from collections.abc import Awaitable, Set
from datetime import timedelta
from typing import Any

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import ComponentCategory

from ..._internal._asyncio import cancel_and_await
from ..._internal._channels import ChannelRegistry, ReceiverFetcher
from ...microgrid import connection_manager
from ...microgrid._data_sourcing import ComponentMetricRequest
from ...microgrid._power_distributing import Result
from ...microgrid._power_distributing._component_status import ComponentPoolStatus
from ...microgrid._power_managing._base_classes import Proposal, ReportRequest
from ..formula_engine._formula_engine_pool import FormulaEnginePool
from ._methods import MetricAggregator


class BatteryPoolReferenceStore:  # pylint: disable=too-many-instance-attributes
    """A class for maintaining the shared state/tasks for a set of pool of batteries.

    This includes ownership of
    - the formula engine pool and metric calculators.
    - the tasks for updating the battery status for the metric calculators.

    These are independent of the priority of the actors and can be shared between
    multiple users of the same set of batteries.

    They are exposed through the BatteryPool class.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
        batteries_status_receiver: Receiver[ComponentPoolStatus],
        power_manager_requests_sender: Sender[Proposal],
        power_manager_bounds_subscription_sender: Sender[ReportRequest],
        power_distribution_results_fetcher: ReceiverFetcher[Result],
        min_update_interval: timedelta,
        batteries_id: Set[int] | None = None,
    ) -> None:
        """Create the class instance.

        Args:
            channel_registry: A channel registry instance shared with the resampling
                actor.
            resampler_subscription_sender: A sender for sending metric requests to the
                resampling actor.
            batteries_status_receiver: Receiver to receive status of the batteries.
                Receivers should has maxsize = 1 to fetch only the latest status.
                Battery status channel should has resend_latest = True.
                It should send information when any battery changed status.
                Battery status should include status of the inverter adjacent to this
                battery.
            power_manager_requests_sender: A Channel sender for sending power
                requests to the power managing actor.
            power_manager_bounds_subscription_sender: A Channel sender for sending
                power bounds requests to the power managing actor.
            power_distribution_results_fetcher: A ReceiverFetcher for the results from
                the power distributing actor.
            min_update_interval: Some metrics in BatteryPool are send only when they
                change. For these metrics min_update_interval is the minimum time
                interval between the following messages.
                Note that this argument is similar to the resampling period
                argument in the ComponentMetricsResamplingActor. But as opposed to
                ResamplingActor, timestamp returned in the resulting message will be
                the timestamp of the last received component data.
                It is currently impossible to use resampling actor for these metrics,
                because we can't specify resampling function for them.
            batteries_id: Subset of the batteries that should be included in the
                battery pool. If None or empty, then all batteries from the microgrid
                will be used.
        """
        self._batteries: frozenset[int]
        if batteries_id:
            self._batteries = frozenset(batteries_id)
        else:
            self._batteries = self._get_all_batteries()

        self._working_batteries: set[int] = set()

        self._update_battery_status_task: asyncio.Task[None] | None = None
        if self._batteries:
            self._update_battery_status_task = asyncio.create_task(
                self._update_battery_status(batteries_status_receiver)
            )

        self._min_update_interval: timedelta = min_update_interval

        self._power_manager_requests_sender: Sender[Proposal] = (
            power_manager_requests_sender
        )

        self._power_manager_bounds_subscription_sender: Sender[ReportRequest] = (
            power_manager_bounds_subscription_sender
        )

        self._active_methods: dict[str, MetricAggregator[Any]] = {}
        self._power_bounds_subs: dict[str, asyncio.Task[None]] = {}
        self._namespace: str = f"battery-pool-{self._batteries}-{uuid.uuid4()}"
        self._power_distributing_namespace: str = f"power-distributor-{self._namespace}"
        self._channel_registry: ChannelRegistry = channel_registry
        self._power_dist_results_fetcher: ReceiverFetcher[Result] = (
            power_distribution_results_fetcher
        )
        self._formula_pool: FormulaEnginePool = FormulaEnginePool(
            self._namespace,
            self._channel_registry,
            resampler_subscription_sender,
        )

    async def stop(self) -> None:
        """Stop all pending async tasks."""
        tasks_to_stop: list[Awaitable[Any]] = [
            method.stop() for method in self._active_methods.values()
        ]
        tasks_to_stop.append(self._formula_pool.stop())
        if self._update_battery_status_task:
            tasks_to_stop.append(cancel_and_await(self._update_battery_status_task))
        await asyncio.gather(*tasks_to_stop)

    def _get_all_batteries(self) -> frozenset[int]:
        """Get all batteries from the microgrid.

        Returns:
            All batteries in the microgrid.
        """
        graph = connection_manager.get().component_graph
        return frozenset(
            {
                battery.component_id
                for battery in graph.components(
                    component_categories={ComponentCategory.BATTERY}
                )
            }
        )

    async def _update_battery_status(
        self, receiver: Receiver[ComponentPoolStatus]
    ) -> None:
        async for status in receiver:
            self._working_batteries = status.get_working_components(self._batteries)
            for item in self._active_methods.values():
                item.update_working_batteries(self._working_batteries)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Steps for building formula engines with."""

from __future__ import annotations

import logging
import math
from abc import ABC, abstractmethod
from typing import Any, Generic

from frequenz.channels import Receiver, ReceiverError

from .._base_types import QuantityT, Sample

_logger = logging.getLogger(__name__)


class FormulaStep(ABC):
    """Represents an individual step/stage in a formula.

    Each step, when applied on to an evaluation stack, would pop its input parameters
    from the stack and push its result back in.
    """

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """

    @abstractmethod
    def apply(self, eval_stack: list[float]) -> None:
        """Apply a formula operation on the eval_stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """


class Adder(FormulaStep):
    """A formula step for adding two values."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "+"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, add them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 + val2
        eval_stack.append(res)


class Subtractor(FormulaStep):
    """A formula step for subtracting one value from another."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "-"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, subtract them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 - val2
        eval_stack.append(res)


class Multiplier(FormulaStep):
    """A formula step for multiplying two values."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "*"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, multiply them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 * val2
        eval_stack.append(res)


class Divider(FormulaStep):
    """A formula step for dividing one value by another."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "/"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack, divide them, push the result back in.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = val1 / val2
        eval_stack.append(res)


class Maximizer(FormulaStep):
    """A formula step that represents the max function."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "max"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack and pushes back the maximum.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = max(val1, val2)
        eval_stack.append(res)


class Minimizer(FormulaStep):
    """A formula step that represents the min function."""

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "min"

    def apply(self, eval_stack: list[float]) -> None:
        """Extract two values from the stack and pushes back the minimum.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val2 = eval_stack.pop()
        val1 = eval_stack.pop()
        res = min(val1, val2)
        eval_stack.append(res)


class Consumption(FormulaStep):
    """A formula step that represents the consumption operator.

    The consumption operator is the maximum of the value on top
    of the evaluation stack and 0.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "consumption"

    def apply(self, eval_stack: list[float]) -> None:
        """
        Apply the consumption formula.

        Replace the top of the eval eval_stack with the same value if the value
        is positive or 0.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        eval_stack.append(max(val, 0))


class Production(FormulaStep):
    """A formula step that represents the production operator.

    The production operator is the maximum of the value times minus one on top
    of the evaluation stack and 0.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "production"

    def apply(self, eval_stack: list[float]) -> None:
        """
        Apply the production formula.

        Replace the top of the eval eval_stack with its absolute value if the
        value is negative or 0.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        eval_stack.append(max(-val, 0))


class OpenParen(FormulaStep):
    """A no-op formula step used while building a prefix formula engine.

    Any OpenParen steps would get removed once a formula is built.
    """

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return "("

    def apply(self, _: list[float]) -> None:
        """No-op."""


class ConstantValue(FormulaStep):
    """A formula step for inserting a constant value."""

    def __init__(self, value: float) -> None:
        """Create a `ConstantValue` instance.

        Args:
            value: The constant value.
        """
        self._value = value

    @property
    def value(self) -> float:
        """Return the constant value.

        Returns:
            The constant value.
        """
        return self._value

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return str(self._value)

    def apply(self, eval_stack: list[float]) -> None:
        """Push the constant value to the eval_stack.

        Args:
            eval_stack: An evaluation stack, to append the constant value to.
        """
        eval_stack.append(self._value)


class Clipper(FormulaStep):
    """A formula step for clipping a value between a minimum and maximum."""

    def __init__(self, min_val: float | None, max_val: float | None) -> None:
        """Create a `Clipper` instance.

        Args:
            min_val: The minimum value.
            max_val: The maximum value.
        """
        self._min_val = min_val
        self._max_val = max_val

    @property
    def min_value(self) -> float | None:
        """Return the minimum value.

        Returns:
            The minimum value.
        """
        return self._min_val

    @property
    def max_value(self) -> float | None:
        """Return the maximum value.

        Returns:
            The maximum value.
        """
        return self._max_val

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return f"clip({self._min_val}, {self._max_val})"

    def apply(self, eval_stack: list[float]) -> None:
        """Clip the value at the top of the eval_stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.
        """
        val = eval_stack.pop()
        if self._min_val is not None:
            val = max(val, self._min_val)
        if self._max_val is not None:
            val = min(val, self._max_val)
        eval_stack.append(val)


class FallbackMetricFetcher(Receiver[Sample[QuantityT]], Generic[QuantityT]):
    """A fallback metric fetcher for formula engines.

    Generates a metric value from the fallback components if the primary metric
    is invalid.

    This class starts running when the primary MetricFetcher starts receiving invalid data.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Get the name of the fetcher."""

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """Check whether the metric fetcher is running."""

    @abstractmethod
    def start(self) -> None:
        """Initialize the metric fetcher and start fetching samples."""


class MetricFetcher(Generic[QuantityT], FormulaStep):
    """A formula step for fetching a value from a metric Receiver."""

    def __init__(
        self,
        name: str,
        stream: Receiver[Sample[QuantityT]],
        *,
        nones_are_zeros: bool,
        fallback: FallbackMetricFetcher[QuantityT] | None = None,
    ) -> None:
        """Create a `MetricFetcher` instance.

        Args:
            name: The name of the metric.
            stream: A channel receiver from which to fetch samples.
            nones_are_zeros: Whether to treat None values from the stream as 0s.
            fallback: Metric fetcher to use if primary one start sending
                invalid data (e.g. due to a component stop). If None, the data from
                primary metric fetcher will be used.
        """
        self._name = name
        self._stream: Receiver[Sample[QuantityT]] = stream
        self._next_value: Sample[QuantityT] | None = None
        self._nones_are_zeros = nones_are_zeros
        self._fallback: FallbackMetricFetcher[QuantityT] | None = fallback
        self._latest_fallback_sample: Sample[QuantityT] | None = None

    @property
    def stream(self) -> Receiver[Sample[QuantityT]]:
        """Return the stream from which to fetch values.

        Returns:
            The stream from which to fetch values.
        """
        return self._stream

    def stream_name(self) -> str:
        """Return the name of the stream.

        Returns:
            The name of the stream.
        """
        return str(self._stream.__doc__)

    def _is_value_valid(self, value: QuantityT | None) -> bool:
        return not (value is None or value.isnan() or value.isinf())

    async def _synchronize_and_fetch_fallback(
        self,
        primary_fetcher_sample: Sample[QuantityT],
        fallback_fetcher: FallbackMetricFetcher[QuantityT],
    ) -> Sample[QuantityT] | None:
        """Synchronize the fallback fetcher and return the fallback value.

        Args:
            primary_fetcher_sample: The sample fetched from the primary fetcher.
            fallback_fetcher: The fallback metric fetcher.

        Returns:
            The value from the synchronized stream. Returns None if the primary
            fetcher sample is older than the latest sample from the fallback
            fetcher or if the fallback fetcher fails to fetch the next value.
        """
        # fallback_fetcher was not used, yet. We need to fetch first value.
        if self._latest_fallback_sample is None:
            try:
                self._latest_fallback_sample = await fallback_fetcher.receive()
            except ReceiverError[Any] as err:
                _logger.error(
                    "Fallback metric fetcher %s failed to fetch next value: %s."
                    "Using primary metric fetcher.",
                    fallback_fetcher.name,
                    err,
                )
                return None

        if primary_fetcher_sample.timestamp < self._latest_fallback_sample.timestamp:
            return None

        # Synchronize the fallback fetcher with primary one
        while primary_fetcher_sample.timestamp > self._latest_fallback_sample.timestamp:
            try:
                self._latest_fallback_sample = await fallback_fetcher.receive()
            except ReceiverError[Any] as err:
                _logger.error(
                    "Fallback metric fetcher %s failed to fetch next value: %s."
                    "Using primary metric fetcher.",
                    fallback_fetcher.name,
                    err,
                )
                return None

        return self._latest_fallback_sample

    async def fetch_next_with_fallback(
        self, fallback_fetcher: FallbackMetricFetcher[QuantityT]
    ) -> Sample[QuantityT]:
        """Fetch the next value from the primary and fallback streams.

        Return the value from the stream that returns a valid value.
        If any stream raises an exception, then return the value from
        the other stream.

        Args:
            fallback_fetcher: The fallback metric fetcher.

        Returns:
            The value fetched from either the primary or fallback stream.
        """
        try:
            primary = await self._stream.receive()
        except ReceiverError[Any] as err:
            _logger.error(
                "Primary metric fetcher %s failed to fetch next value: %s."
                "Using fallback metric fetcher.",
                self._name,
                err,
            )
            return await fallback_fetcher.receive()

        fallback = await self._synchronize_and_fetch_fallback(primary, fallback_fetcher)
        if fallback is None:
            return primary

        if self._is_value_valid(primary.value):
            return primary
        return fallback

    async def fetch_next(self) -> Sample[QuantityT] | None:
        """Fetch the next value from the stream.

        To be called before each call to `apply`.

        Returns:
            The fetched Sample.
        """
        self._next_value = await self._fetch_next()
        return self._next_value

    async def _fetch_next(self) -> Sample[QuantityT] | None:
        if self._fallback is None:
            return await self._stream.receive()

        if self._fallback.is_running:
            return await self.fetch_next_with_fallback(self._fallback)

        next_value = None
        try:
            next_value = await self._stream.receive()
        except ReceiverError[Any] as err:
            _logger.error("Failed to fetch next value from %s: %s", self._name, err)
        else:
            if self._is_value_valid(next_value.value):
                return next_value

        _logger.warning(
            "Primary metric %s is invalid. Running fallback metric fetcher: %s",
            self._name,
            self._fallback.name,
        )
        # start fallback formula but don't wait for it because it has to
        # synchronize. Just return invalid value.
        self._fallback.start()
        return next_value

    @property
    def value(self) -> Sample[QuantityT] | None:
        """Get the next value in the stream.

        Returns:
            Next value in the stream.
        """
        return self._next_value

    def __repr__(self) -> str:
        """Return a string representation of the step.

        Returns:
            A string representation of the step.
        """
        return self._name

    def apply(self, eval_stack: list[float]) -> None:
        """Push the latest value from the stream into the evaluation stack.

        Args:
            eval_stack: An evaluation stack, to apply the formula step on.

        Raises:
            RuntimeError: No next value available to append.
        """
        if self._next_value is None:
            raise RuntimeError("No next value available to append.")

        next_value = self._next_value.value
        if next_value is None or next_value.isnan() or next_value.isinf():
            if self._nones_are_zeros:
                eval_stack.append(0.0)
            else:
                eval_stack.append(math.nan)
        else:
            eval_stack.append(next_value.base_value)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""An actor to resample microgrid component metrics."""


import asyncio
import dataclasses
import logging

from frequenz.channels import Receiver, Sender
from frequenz.quantities import Quantity

from .._internal._asyncio import cancel_and_await
from .._internal._channels import ChannelRegistry
from ..actor import Actor
from ..timeseries import Sample
from ..timeseries._resampling import Resampler, ResamplerConfig, ResamplingError
from ._data_sourcing import ComponentMetricRequest

_logger = logging.getLogger(__name__)


class ComponentMetricsResamplingActor(Actor):
    """An actor to resample microgrid component metrics."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        channel_registry: ChannelRegistry,
        data_sourcing_request_sender: Sender[ComponentMetricRequest],
        resampling_request_receiver: Receiver[ComponentMetricRequest],
        config: ResamplerConfig,
        name: str | None = None,
    ) -> None:
        """Initialize an instance.

        Args:
            channel_registry: The channel registry used to get senders and
                receivers for data sourcing subscriptions.
            data_sourcing_request_sender: The sender used to send requests to
                the [`DataSourcingActor`][frequenz.sdk.actor.DataSourcingActor]
                to subscribe to component metrics.
            resampling_request_receiver: The receiver to use to receive new
                resampling subscription requests.
            config: The configuration for the resampler.
            name: The name of the actor. If `None`, `str(id(self))` will be used. This
                is used mostly for debugging purposes.
        """
        super().__init__(name=name)
        self._channel_registry: ChannelRegistry = channel_registry
        self._data_sourcing_request_sender: Sender[ComponentMetricRequest] = (
            data_sourcing_request_sender
        )
        self._resampling_request_receiver: Receiver[ComponentMetricRequest] = (
            resampling_request_receiver
        )
        self._resampler: Resampler = Resampler(config)
        self._active_req_channels: set[str] = set()

    async def _subscribe(self, request: ComponentMetricRequest) -> None:
        """Request data for a component metric.

        Args:
            request: The request for component metric data.
        """
        request_channel_name = request.get_channel_name()

        # If we are already handling this request, there is nothing to do.
        if request_channel_name in self._active_req_channels:
            return

        self._active_req_channels.add(request_channel_name)

        data_source_request = dataclasses.replace(
            request, namespace=request.namespace + ":Source"
        )
        data_source_channel_name = data_source_request.get_channel_name()
        await self._data_sourcing_request_sender.send(data_source_request)
        receiver = self._channel_registry.get_or_create(
            Sample[Quantity], data_source_channel_name
        ).new_receiver()

        # This is a temporary hack until the Sender implementation uses
        # exceptions to report errors.
        sender = self._channel_registry.get_or_create(
            Sample[Quantity], request_channel_name
        ).new_sender()

        self._resampler.add_timeseries(request_channel_name, receiver, sender.send)

    async def _process_resampling_requests(self) -> None:
        """Process resampling data requests."""
        async for request in self._resampling_request_receiver:
            await self._subscribe(request)

    async def _run(self) -> None:
        """Resample known component metrics and process resampling requests.

        If there is a resampling error while resampling some component metric,
        then that metric will be discarded and not resampled any more. Any
        other error will be propagated (most likely ending in the actor being
        restarted).

        This method creates 2 main tasks:

        - One task to process incoming subscription requests to resample new metrics.
        - One task to run the resampler.
        """
        tasks_to_cancel: set[asyncio.Task[None]] = set()
        subscriptions_task: asyncio.Task[None] | None = None
        resampling_task: asyncio.Task[None] | None = None

        try:
            while True:
                if subscriptions_task is None or subscriptions_task.done():
                    subscriptions_task = asyncio.create_task(
                        self._process_resampling_requests()
                    )
                    tasks_to_cancel.add(subscriptions_task)

                if resampling_task is None or resampling_task.done():
                    resampling_task = asyncio.create_task(self._resampler.resample())
                    tasks_to_cancel.add(resampling_task)

                done, _ = await asyncio.wait(
                    [resampling_task, subscriptions_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )

                if subscriptions_task in done:
                    tasks_to_cancel.remove(subscriptions_task)
                    self._log_subscriptions_task_error(subscriptions_task)

                if resampling_task in done:
                    tasks_to_cancel.remove(resampling_task)
                    self._log_resampling_task_error(resampling_task)

        finally:
            await asyncio.gather(*[cancel_and_await(t) for t in tasks_to_cancel])

            # XXX: Here we should probably do a:  pylint: disable=fixme
            # await self._resampler.stop()
            # But since the actor will be restarted, the internal state would
            # be broken if we stop the resampler.
            #
            # We have an even bigger problem with this naive restarting
            # approach, as restarting this actor without really resetting its
            # state would be mostly the same as not really leaving the run()
            # method and just swallow any exception, which doesn't look super
            # smart.

    def _log_subscriptions_task_error(
        self, subscriptions_task: asyncio.Task[None]
    ) -> None:
        """Log an error from a stopped subscriptions task.

        Args:
            subscriptions_task: The subscriptions task.
        """
        try:
            subscriptions_task.result()
        # pylint: disable-next=broad-except
        except (Exception, asyncio.CancelledError):
            _logger.exception(
                "The subscriptions task ended with an exception, restarting..."
            )
        else:
            _logger.error("The subscriptions task ended unexpectedly, restarting...")

    def _log_resampling_task_error(self, resampling_task: asyncio.Task[None]) -> None:
        """Log an error from a stopped resampling task.

        Args:
            resampling_task: The resampling task.
        """
        # The resampler shouldn't be cancelled or end without an exception
        try:
            resampling_task.result()
        except ResamplingError as error:
            for source, source_error in error.exceptions.items():
                _logger.error(
                    "Error resampling source %s, removing source",
                    source,
                    exc_info=source_error,
                )
                removed = self._resampler.remove_timeseries(source)
                if not removed:
                    _logger.error(
                        "Got an exception from an unknown source: "
                        "source=%r, exception=%r",
                        source,
                        source_error,
                    )
        # pylint: disable-next=broad-except
        except (Exception, asyncio.CancelledError):
            # We don't know what to do with something other than
            # ResamplingError, so we log it, restart, and hope for the best.
            _logger.exception(
                "The resample() function got an unexpected error, restarting..."
            )
        else:
            # The resample function should not end normally, so we log it,
            # restart, and hope for the best.
            _logger.error(
                "The resample() function ended without an exception, restarting..."
            )



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for 3-phase Grid Current."""


import logging
from collections import abc

from frequenz.client.microgrid import ComponentMetricId
from frequenz.quantities import Current

from .._formula_engine import FormulaEngine, FormulaEngine3Phase
from ._formula_generator import NON_EXISTING_COMPONENT_ID, FormulaGenerator

_logger = logging.getLogger(__name__)


class EVChargerCurrentFormula(FormulaGenerator[Current]):
    """Create a formula engine from the component graph for calculating grid current."""

    def generate(self) -> FormulaEngine3Phase[Current]:
        """Generate a formula for calculating total EV current for given component ids.

        Returns:
            A formula engine that calculates total 3-phase EV Charger current values.
        """
        component_ids = self._config.component_ids

        if not component_ids:
            _logger.warning(
                "No EV Charger component IDs specified. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no EV Chargers, we have to send 0 values as the same
            # frequency as the other streams.  So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder = self._get_builder(
                "ev-current", ComponentMetricId.ACTIVE_POWER, Current.from_amperes
            )
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            engine = builder.build()
            return FormulaEngine3Phase(
                "ev-current",
                Current.from_amperes,
                (engine, engine, engine),
            )

        return FormulaEngine3Phase(
            "ev-current",
            Current.from_amperes,
            (
                (
                    self._gen_phase_formula(
                        component_ids, ComponentMetricId.CURRENT_PHASE_1
                    )
                ),
                (
                    self._gen_phase_formula(
                        component_ids, ComponentMetricId.CURRENT_PHASE_2
                    )
                ),
                (
                    self._gen_phase_formula(
                        component_ids, ComponentMetricId.CURRENT_PHASE_3
                    )
                ),
            ),
        )

    def _gen_phase_formula(
        self,
        component_ids: abc.Set[int],
        metric_id: ComponentMetricId,
    ) -> FormulaEngine[Current]:
        builder = self._get_builder("ev-current", metric_id, Current.from_amperes)

        # generate a formula that just adds values from all EV Chargers.
        for idx, component_id in enumerate(component_ids):
            if idx > 0:
                builder.push_oper("+")

            builder.push_component_metric(component_id, nones_are_zeros=True)

        return builder.build()



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Utilities to deal with configuration."""

from collections.abc import Mapping
from typing import Any, ClassVar, Protocol, TypeVar, cast

from marshmallow import Schema
from marshmallow_dataclass import class_schema


# This is a hack that relies on identifying dataclasses by looking into an undocumented
# property of dataclasses[1], so it might break in the future. Nevertheless, it seems to
# be widely used in the community, for example `mypy` and `pyright` seem to rely on
# it[2].
#
# [1]: https://github.com/python/mypy/issues/15974#issuecomment-1694781006
# [2]: https://github.com/python/mypy/issues/15974#issuecomment-1694993493
class Dataclass(Protocol):
    """A protocol for dataclasses."""

    __dataclass_fields__: ClassVar[dict[str, Any]]
    """The fields of the dataclass."""


DataclassT = TypeVar("DataclassT", bound=Dataclass)
"""Type variable for configuration classes."""


def load_config(
    cls: type[DataclassT],
    config: Mapping[str, Any],
    /,
    base_schema: type[Schema] | None = None,
    **marshmallow_load_kwargs: Any,
) -> DataclassT:
    """Load a configuration from a dictionary into an instance of a configuration class.

    The configuration class is expected to be a [`dataclasses.dataclass`][], which is
    used to create a [`marshmallow.Schema`][] schema to validate the configuration
    dictionary using [`marshmallow_dataclass.class_schema`][] (which in turn uses the
    [`marshmallow.Schema.load`][] method to do the validation and deserialization).

    To customize the schema derived from the configuration dataclass, you can use the
    `metadata` key in [`dataclasses.field`][] to pass extra options to
    [`marshmallow_dataclass`][] to be used during validation and deserialization.

    Additional arguments can be passed to [`marshmallow.Schema.load`][] using keyword
    arguments `marshmallow_load_kwargs`.

    Note:
        This method will raise [`marshmallow.ValidationError`][] if the configuration
        dictionary is invalid and you have to have in mind all of the gotchas of
        [`marshmallow`][] and [`marshmallow_dataclass`][] applies when using this
        function.  It is recommended to carefully read the documentation of these
        libraries.

    Args:
        cls: The configuration class.
        config: The configuration dictionary.
        base_schema: An optional class to be used as a base schema for the configuration
            class. This allow using custom fields for example. Will be passed to
            [`marshmallow_dataclass.class_schema`][].
        **marshmallow_load_kwargs: Additional arguments to be passed to
            [`marshmallow.Schema.load`][].

    Returns:
        The loaded configuration as an instance of the configuration class.
    """
    instance = class_schema(cls, base_schema)().load(config, **marshmallow_load_kwargs)
    # We need to cast because `.load()` comes from marshmallow and doesn't know which
    # type is returned.
    return cast(DataclassT, instance)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Create, connect and own instances of data pipeline components.

Provides SDK users direct access to higher level components of the data pipeline,
eliminating the boiler plate code required to setup the DataSourcingActor and the
ResamplingActor.
"""

from __future__ import annotations

import logging
import typing
from collections import abc
from dataclasses import dataclass
from datetime import timedelta

from frequenz.channels import Broadcast, Sender
from frequenz.client.microgrid import ComponentCategory, InverterType

from .._internal._channels import ChannelRegistry
from ..actor._actor import Actor
from ..timeseries import ResamplerConfig
from ..timeseries._grid_frequency import GridFrequency
from ..timeseries._voltage_streamer import VoltageStreamer
from ..timeseries.grid import Grid
from ..timeseries.grid import get as get_grid
from ..timeseries.grid import initialize as initialize_grid
from ._data_sourcing import ComponentMetricRequest, DataSourcingActor
from ._power_wrapper import PowerWrapper

# A number of imports had to be done inside functions where they are used, to break
# import cycles.
#
# pylint: disable=import-outside-toplevel
if typing.TYPE_CHECKING:
    from ..timeseries.battery_pool import BatteryPool
    from ..timeseries.battery_pool._battery_pool_reference_store import (
        BatteryPoolReferenceStore,
    )
    from ..timeseries.consumer import Consumer
    from ..timeseries.ev_charger_pool import EVChargerPool
    from ..timeseries.ev_charger_pool._ev_charger_pool_reference_store import (
        EVChargerPoolReferenceStore,
    )
    from ..timeseries.logical_meter import LogicalMeter
    from ..timeseries.producer import Producer
    from ..timeseries.pv_pool import PVPool
    from ..timeseries.pv_pool._pv_pool_reference_store import PVPoolReferenceStore

_logger = logging.getLogger(__name__)


_REQUEST_RECV_BUFFER_SIZE = 500
"""The maximum number of requests that can be queued in the request receiver.

A larger buffer size means that the DataSourcing and Resampling actors don't drop
requests and will be able to keep up with higher request rates in larger installations.
"""


@dataclass
class _ActorInfo:
    """Holds instances of core data pipeline actors and their request channels."""

    actor: Actor
    """The actor instance."""

    channel: Broadcast[ComponentMetricRequest]
    """The request channel for the actor."""


class _DataPipeline:  # pylint: disable=too-many-instance-attributes
    """Create, connect and own instances of data pipeline components.

    Provides SDK users direct access to higher level components of the data pipeline,
    eliminating the boiler plate code required to setup the DataSourcingActor and the
    ResamplingActor.
    """

    def __init__(
        self,
        resampler_config: ResamplerConfig,
        api_power_request_timeout: timedelta = timedelta(seconds=5.0),
    ) -> None:
        """Create a `DataPipeline` instance.

        Args:
            resampler_config: Config to pass on to the resampler.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
        """
        self._resampler_config: ResamplerConfig = resampler_config

        self._channel_registry: ChannelRegistry = ChannelRegistry(
            name="Data Pipeline Registry"
        )

        self._data_sourcing_actor: _ActorInfo | None = None
        self._resampling_actor: _ActorInfo | None = None

        self._battery_power_wrapper = PowerWrapper(
            self._channel_registry,
            api_power_request_timeout=api_power_request_timeout,
            component_category=ComponentCategory.BATTERY,
        )
        self._ev_power_wrapper = PowerWrapper(
            self._channel_registry,
            api_power_request_timeout=api_power_request_timeout,
            component_category=ComponentCategory.EV_CHARGER,
        )
        self._pv_power_wrapper = PowerWrapper(
            self._channel_registry,
            api_power_request_timeout=api_power_request_timeout,
            component_category=ComponentCategory.INVERTER,
            component_type=InverterType.SOLAR,
        )

        self._logical_meter: LogicalMeter | None = None
        self._consumer: Consumer | None = None
        self._producer: Producer | None = None
        self._grid: Grid | None = None
        self._ev_charger_pool_reference_stores: dict[
            frozenset[int], EVChargerPoolReferenceStore
        ] = {}
        self._battery_pool_reference_stores: dict[
            frozenset[int], BatteryPoolReferenceStore
        ] = {}
        self._pv_pool_reference_stores: dict[frozenset[int], PVPoolReferenceStore] = {}
        self._frequency_instance: GridFrequency | None = None
        self._voltage_instance: VoltageStreamer | None = None

        self._known_pool_keys: set[str] = set()
        """A set of keys for corresponding to created EVChargerPool instances.

        This is used to warn the user if they try to create a new EVChargerPool instance
        for the same set of component IDs, and with the same priority.
        """

    def frequency(self) -> GridFrequency:
        """Return the grid frequency measuring point."""
        if self._frequency_instance is None:
            self._frequency_instance = GridFrequency(
                self._data_sourcing_request_sender(),
                self._channel_registry,
            )

        return self._frequency_instance

    def voltage_per_phase(self) -> VoltageStreamer:
        """Return the per-phase voltage measuring point."""
        if not self._voltage_instance:
            self._voltage_instance = VoltageStreamer(
                self._resampling_request_sender(),
                self._channel_registry,
            )

        return self._voltage_instance

    def logical_meter(self) -> LogicalMeter:
        """Return the logical meter of the microgrid."""
        from ..timeseries.logical_meter import LogicalMeter

        if self._logical_meter is None:
            self._logical_meter = LogicalMeter(
                channel_registry=self._channel_registry,
                resampler_subscription_sender=self._resampling_request_sender(),
            )
        return self._logical_meter

    def consumer(self) -> Consumer:
        """Return the consumption measuring point of the microgrid."""
        from ..timeseries.consumer import Consumer

        if self._consumer is None:
            self._consumer = Consumer(
                channel_registry=self._channel_registry,
                resampler_subscription_sender=self._resampling_request_sender(),
            )
        return self._consumer

    def producer(self) -> Producer:
        """Return the production measuring point of the microgrid."""
        from ..timeseries.producer import Producer

        if self._producer is None:
            self._producer = Producer(
                channel_registry=self._channel_registry,
                resampler_subscription_sender=self._resampling_request_sender(),
            )
        return self._producer

    def grid(self) -> Grid:
        """Return the grid measuring point."""
        if self._grid is None:
            initialize_grid(
                channel_registry=self._channel_registry,
                resampler_subscription_sender=self._resampling_request_sender(),
            )
            self._grid = get_grid()

        return self._grid

    def new_ev_charger_pool(
        self,
        *,
        priority: int,
        component_ids: abc.Set[int] | None = None,
        name: str | None = None,
        set_operating_point: bool = False,
    ) -> EVChargerPool:
        """Return the corresponding EVChargerPool instance for the given ids.

        If an EVChargerPool instance for the given ids doesn't exist, a new one is
        created and returned.

        Args:
            priority: The priority of the actor making the call.
            component_ids: Optional set of IDs of EV Chargers to be managed by the
                EVChargerPool.
            name: An optional name used to identify this instance of the pool or a
                corresponding actor in the logs.
            set_operating_point: Whether this instance sets the operating point power or
                the normal power for the components.

        Returns:
            An EVChargerPool instance.
        """
        from ..timeseries.ev_charger_pool import EVChargerPool
        from ..timeseries.ev_charger_pool._ev_charger_pool_reference_store import (
            EVChargerPoolReferenceStore,
        )

        if not self._ev_power_wrapper.started:
            self._ev_power_wrapper.start()

        # We use frozenset to make a hashable key from the input set.
        ref_store_key: frozenset[int] = frozenset()
        if component_ids is not None:
            ref_store_key = frozenset(component_ids)

        pool_key = f"{ref_store_key}-{priority}"
        if pool_key in self._known_pool_keys:
            _logger.warning(
                "An EVChargerPool instance was already created for ev_charger_ids=%s "
                "and priority=%s using `microgrid.ev_charger_pool(...)`."
                "\n  Hint: If the multiple instances are created from the same actor, "
                "consider reusing the same instance."
                "\n  Hint: If the instances are created from different actors, "
                "consider using different priorities to distinguish them.",
                component_ids,
                priority,
            )
        else:
            self._known_pool_keys.add(pool_key)

        if ref_store_key not in self._ev_charger_pool_reference_stores:
            self._ev_charger_pool_reference_stores[ref_store_key] = (
                EVChargerPoolReferenceStore(
                    channel_registry=self._channel_registry,
                    resampler_subscription_sender=self._resampling_request_sender(),
                    status_receiver=self._ev_power_wrapper.status_channel.new_receiver(
                        limit=1
                    ),
                    power_manager_requests_sender=(
                        self._ev_power_wrapper.proposal_channel.new_sender()
                    ),
                    power_manager_bounds_subs_sender=(
                        self._ev_power_wrapper.bounds_subscription_channel.new_sender()
                    ),
                    power_distribution_results_fetcher=(
                        self._ev_power_wrapper.distribution_results_fetcher()
                    ),
                    component_ids=component_ids,
                )
            )
        return EVChargerPool(
            pool_ref_store=self._ev_charger_pool_reference_stores[ref_store_key],
            name=name,
            priority=priority,
            set_operating_point=set_operating_point,
        )

    def new_pv_pool(
        self,
        *,
        priority: int,
        component_ids: abc.Set[int] | None = None,
        name: str | None = None,
        set_operating_point: bool = False,
    ) -> PVPool:
        """Return a new `PVPool` instance for the given ids.

        If a `PVPoolReferenceStore` instance for the given PV inverter ids doesn't
        exist, a new one is created and used for creating the `PVPool`.

        Args:
            priority: The priority of the actor making the call.
            component_ids: Optional set of IDs of PV inverters to be managed by the
                `PVPool`.
            name: An optional name used to identify this instance of the pool or a
                corresponding actor in the logs.
            set_operating_point: Whether this instance sets the operating point power or
                the normal power for the components.

        Returns:
            A `PVPool` instance.
        """
        from ..timeseries.pv_pool import PVPool
        from ..timeseries.pv_pool._pv_pool_reference_store import PVPoolReferenceStore

        if not self._pv_power_wrapper.started:
            self._pv_power_wrapper.start()

        # We use frozenset to make a hashable key from the input set.
        ref_store_key: frozenset[int] = frozenset()
        if component_ids is not None:
            ref_store_key = frozenset(component_ids)

        pool_key = f"{ref_store_key}-{priority}"
        if pool_key in self._known_pool_keys:
            _logger.warning(
                "A PVPool instance was already created for pv_inverter_ids=%s and "
                "priority=%s using `microgrid.pv_pool(...)`."
                "\n  Hint: If the multiple instances are created from the same actor, "
                "consider reusing the same instance."
                "\n  Hint: If the instances are created from different actors, "
                "consider using different priorities to distinguish them.",
                component_ids,
                priority,
            )
        else:
            self._known_pool_keys.add(pool_key)

        if ref_store_key not in self._pv_pool_reference_stores:
            self._pv_pool_reference_stores[ref_store_key] = PVPoolReferenceStore(
                channel_registry=self._channel_registry,
                resampler_subscription_sender=self._resampling_request_sender(),
                status_receiver=(
                    self._pv_power_wrapper.status_channel.new_receiver(limit=1)
                ),
                power_manager_requests_sender=(
                    self._pv_power_wrapper.proposal_channel.new_sender()
                ),
                power_manager_bounds_subs_sender=(
                    self._pv_power_wrapper.bounds_subscription_channel.new_sender()
                ),
                power_distribution_results_fetcher=(
                    self._pv_power_wrapper.distribution_results_fetcher()
                ),
                component_ids=component_ids,
            )

        return PVPool(
            pool_ref_store=self._pv_pool_reference_stores[ref_store_key],
            name=name,
            priority=priority,
            set_operating_point=set_operating_point,
        )

    def new_battery_pool(
        self,
        *,
        priority: int,
        component_ids: abc.Set[int] | None = None,
        name: str | None = None,
        set_operating_point: bool = False,
    ) -> BatteryPool:
        """Return a new `BatteryPool` instance for the given ids.

        If a `BatteryPoolReferenceStore` instance for the given battery ids doesn't
        exist, a new one is created and used for creating the `BatteryPool`.

        Args:
            priority: The priority of the actor making the call.
            component_ids: Optional set of IDs of batteries to be managed by the
                `BatteryPool`.
            name: An optional name used to identify this instance of the pool or a
                corresponding actor in the logs.
            set_operating_point: Whether this instance sets the operating point power or
                the normal power for the components.

        Returns:
            A `BatteryPool` instance.
        """
        from ..timeseries.battery_pool import BatteryPool
        from ..timeseries.battery_pool._battery_pool_reference_store import (
            BatteryPoolReferenceStore,
        )

        if not self._battery_power_wrapper.started:
            self._battery_power_wrapper.start()

        # We use frozenset to make a hashable key from the input set.
        ref_store_key: frozenset[int] = frozenset()
        if component_ids is not None:
            ref_store_key = frozenset(component_ids)

        pool_key = f"{ref_store_key}-{priority}"
        if pool_key in self._known_pool_keys:
            _logger.warning(
                "A BatteryPool instance was already created for battery_ids=%s and "
                "priority=%s using `microgrid.battery_pool(...)`."
                "\n  Hint: If the multiple instances are created from the same actor, "
                "consider reusing the same instance."
                "\n  Hint: If the instances are created from different actors, "
                "consider using different priorities to distinguish them.",
                component_ids,
                priority,
            )
        else:
            self._known_pool_keys.add(pool_key)

        if ref_store_key not in self._battery_pool_reference_stores:
            self._battery_pool_reference_stores[ref_store_key] = (
                BatteryPoolReferenceStore(
                    channel_registry=self._channel_registry,
                    resampler_subscription_sender=self._resampling_request_sender(),
                    batteries_status_receiver=(
                        self._battery_power_wrapper.status_channel.new_receiver(limit=1)
                    ),
                    power_manager_requests_sender=(
                        self._battery_power_wrapper.proposal_channel.new_sender()
                    ),
                    power_manager_bounds_subscription_sender=(
                        self._battery_power_wrapper.bounds_subscription_channel.new_sender()
                    ),
                    power_distribution_results_fetcher=(
                        self._battery_power_wrapper.distribution_results_fetcher()
                    ),
                    min_update_interval=self._resampler_config.resampling_period,
                    batteries_id=component_ids,
                )
            )

        return BatteryPool(
            pool_ref_store=self._battery_pool_reference_stores[ref_store_key],
            name=name,
            priority=priority,
            set_operating_point=set_operating_point,
        )

    def _data_sourcing_request_sender(self) -> Sender[ComponentMetricRequest]:
        """Return a Sender for sending requests to the data sourcing actor.

        If the data sourcing actor is not already running, this function also starts it.

        Returns:
            A Sender for sending requests to the data sourcing actor.
        """
        if self._data_sourcing_actor is None:
            channel: Broadcast[ComponentMetricRequest] = Broadcast(
                name="Data Pipeline: Data Sourcing Actor Request Channel"
            )
            actor = DataSourcingActor(
                request_receiver=channel.new_receiver(limit=_REQUEST_RECV_BUFFER_SIZE),
                registry=self._channel_registry,
            )
            self._data_sourcing_actor = _ActorInfo(actor, channel)
            self._data_sourcing_actor.actor.start()
        return self._data_sourcing_actor.channel.new_sender()

    def _resampling_request_sender(self) -> Sender[ComponentMetricRequest]:
        """Return a Sender for sending requests to the resampling actor.

        If the resampling actor is not already running, this function also starts it.

        Returns:
            A Sender for sending requests to the resampling actor.
        """
        from ._resampling import ComponentMetricsResamplingActor

        if self._resampling_actor is None:
            channel: Broadcast[ComponentMetricRequest] = Broadcast(
                name="Data Pipeline: Component Metric Resampling Actor Request Channel"
            )
            actor = ComponentMetricsResamplingActor(
                channel_registry=self._channel_registry,
                data_sourcing_request_sender=self._data_sourcing_request_sender(),
                resampling_request_receiver=channel.new_receiver(
                    limit=_REQUEST_RECV_BUFFER_SIZE,
                    name=channel.name + " Receiver",
                ),
                config=self._resampler_config,
            )
            self._resampling_actor = _ActorInfo(actor, channel)
            self._resampling_actor.actor.start()
        return self._resampling_actor.channel.new_sender()

    async def _stop(self) -> None:
        """Stop the data pipeline actors."""
        if self._data_sourcing_actor:
            await self._data_sourcing_actor.actor.stop()
        if self._resampling_actor:
            await self._resampling_actor.actor.stop()
        await self._battery_power_wrapper.stop()
        for pool in self._battery_pool_reference_stores.values():
            await pool.stop()


_DATA_PIPELINE: _DataPipeline | None = None


async def initialize(
    resampler_config: ResamplerConfig,
    api_power_request_timeout: timedelta = timedelta(seconds=5.0),
) -> None:
    """Initialize a `DataPipeline` instance.

    Args:
        resampler_config: Config to pass on to the resampler.
        api_power_request_timeout: Timeout to use when making power requests to
            the microgrid API.  When requests to components timeout, they will
            be marked as blocked for a short duration, during which time they
            will be unavailable from the corresponding component pools.

    Raises:
        RuntimeError: if the DataPipeline is already initialized.
    """
    global _DATA_PIPELINE  # pylint: disable=global-statement

    if _DATA_PIPELINE is not None:
        raise RuntimeError("DataPipeline is already initialized.")
    _DATA_PIPELINE = _DataPipeline(resampler_config, api_power_request_timeout)


def frequency() -> GridFrequency:
    """Return the grid frequency measuring point."""
    return _get().frequency()


def voltage_per_phase() -> VoltageStreamer:
    """Return the per-phase voltage measuring point."""
    return _get().voltage_per_phase()


def logical_meter() -> LogicalMeter:
    """Return the logical meter of the microgrid."""
    return _get().logical_meter()


def consumer() -> Consumer:
    """Return the [`Consumption`][frequenz.sdk.timeseries.consumer.Consumer] measuring point."""
    return _get().consumer()


def producer() -> Producer:
    """Return the [`Production`][frequenz.sdk.timeseries.producer.Producer] measuring point."""
    return _get().producer()


def new_ev_charger_pool(
    *,
    priority: int,
    component_ids: abc.Set[int] | None = None,
    name: str | None = None,
    set_operating_point: bool = False,
) -> EVChargerPool:
    """Return a new `EVChargerPool` instance for the given parameters.

    The priority value is used to resolve conflicts when multiple actors are trying to
    propose different power values for the same set of EV chargers.

    !!! note
        When specifying priority, bigger values indicate higher priority.

        It is recommended to reuse the same instance of the `EVChargerPool` within the
        same actor, unless they are managing different sets of EV chargers.

        In deployments with multiple actors managing the same set of EV chargers, it is
        recommended to use different priorities to distinguish between them.  If not,
        a random prioritization will be imposed on them to resolve conflicts, which may
        lead to unexpected behavior like longer duration to converge on the desired
        power.

    Args:
        priority: The priority of the actor making the call.
        component_ids: Optional set of IDs of EV Chargers to be managed by the
            EVChargerPool.  If not specified, all EV Chargers available in the
            component graph are used.
        name: An optional name used to identify this instance of the pool or a
            corresponding actor in the logs.
        set_operating_point: Whether this instance sets the operating point power or the
            normal power for the components.

    Returns:
        An `EVChargerPool` instance.
    """
    return _get().new_ev_charger_pool(
        priority=priority,
        component_ids=component_ids,
        name=name,
        set_operating_point=set_operating_point,
    )


def new_battery_pool(
    *,
    priority: int,
    component_ids: abc.Set[int] | None = None,
    name: str | None = None,
    set_operating_point: bool = False,
) -> BatteryPool:
    """Return a new `BatteryPool` instance for the given parameters.

    The priority value is used to resolve conflicts when multiple actors are trying to
    propose different power values for the same set of batteries.

    !!! note
        When specifying priority, bigger values indicate higher priority.

        It is recommended to reuse the same instance of the `BatteryPool` within the
        same actor, unless they are managing different sets of batteries.

        In deployments with multiple actors managing the same set of batteries, it is
        recommended to use different priorities to distinguish between them.  If not,
        a random prioritization will be imposed on them to resolve conflicts, which may
        lead to unexpected behavior like longer duration to converge on the desired
        power.

    Args:
        priority: The priority of the actor making the call.
        component_ids: Optional set of IDs of batteries to be managed by the
            `BatteryPool`.  If not specified, all batteries available in the component
            graph are used.
        name: An optional name used to identify this instance of the pool or a
            corresponding actor in the logs.
        set_operating_point: Whether this instance sets the operating point power or the
            normal power for the components.

    Returns:
        A `BatteryPool` instance.
    """
    return _get().new_battery_pool(
        priority=priority,
        component_ids=component_ids,
        name=name,
        set_operating_point=set_operating_point,
    )


def new_pv_pool(
    *,
    priority: int,
    component_ids: abc.Set[int] | None = None,
    name: str | None = None,
    set_operating_point: bool = False,
) -> PVPool:
    """Return a new `PVPool` instance for the given parameters.

    The priority value is used to resolve conflicts when multiple actors are trying to
    propose different power values for the same set of PV inverters.

    !!! note
        When specifying priority, bigger values indicate higher priority.

        It is recommended to reuse the same instance of the `PVPool` within the same
        actor, unless they are managing different sets of PV inverters.

        In deployments with multiple actors managing the same set of PV inverters, it is
        recommended to use different priorities to distinguish between them.  If not,
        a random prioritization will be imposed on them to resolve conflicts, which may
        lead to unexpected behavior like longer duration to converge on the desired
        power.

    Args:
        priority: The priority of the actor making the call.
        component_ids: Optional set of IDs of PV inverters to be managed by the
            `PVPool`. If not specified, all PV inverters available in the component
            graph are used.
        name: An optional name used to identify this instance of the pool or a
            corresponding actor in the logs.
        set_operating_point: Whether this instance sets the operating point power or the
            normal power for the components.

    Returns:
        A `PVPool` instance.
    """
    return _get().new_pv_pool(
        priority=priority,
        component_ids=component_ids,
        name=name,
        set_operating_point=set_operating_point,
    )


def grid() -> Grid:
    """Return the grid measuring point."""
    return _get().grid()


def _get() -> _DataPipeline:
    if _DATA_PIPELINE is None:
        raise RuntimeError(
            "DataPipeline is not initialized. "
            "Call `await microgrid.initialize()` first."
        )
    return _DATA_PIPELINE



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""A data window that moves with the latest datapoints of a data stream."""


import asyncio
import logging
import math
from collections.abc import Sequence
from datetime import datetime, timedelta
from typing import SupportsIndex, assert_never, overload

import numpy as np
from frequenz.channels import Broadcast, Receiver, Sender
from frequenz.quantities import Quantity
from numpy.typing import ArrayLike

from ..actor._background_service import BackgroundService
from ._base_types import UNIX_EPOCH, Sample
from ._resampling import Resampler, ResamplerConfig
from ._ringbuffer import OrderedRingBuffer

_logger = logging.getLogger(__name__)


class MovingWindow(BackgroundService):
    """
    A data window that moves with the latest datapoints of a data stream.

    After initialization the `MovingWindow` can be accessed by an integer
    index or a timestamp. A sub window can be accessed by using a slice of
    integers or timestamps.

    Note that a numpy ndarray is returned and thus users can use
    numpys operations directly on a window.

    The window uses a ring buffer for storage and the first element is aligned to
    a fixed defined point in time. Since the moving nature of the window, the
    date of the first and the last element are constantly changing and therefore
    the point in time that defines the alignment can be outside of the time window.
    Modulo arithmetic is used to move the `align_to` timestamp into the latest
    window.

    If for example the `align_to` parameter is set to
    `datetime(1, 1, 1, tzinfo=timezone.utc)` and the window size is bigger than
    one day then the first element will always be aligned to midnight.

    Resampling might be required to reduce the number of samples to store, and
    it can be set by specifying the resampler config parameter so that the user
    can control the granularity of the samples to be stored in the underlying
    buffer.

    If resampling is not required, the resampler config parameter can be
    set to None in which case the MovingWindow will not perform any resampling.

    Example: Calculate the mean of a time interval

        ```python
        from datetime import datetime, timedelta, timezone

        async def send_mock_data(sender: Sender[Sample]) -> None:
            while True:
                await sender.send(Sample(datetime.now(tz=timezone.utc), 10.0))
                await asyncio.sleep(1.0)

        async def run() -> None:
            resampled_data_channel = Broadcast[Sample](name="sample-data")
            resampled_data_receiver = resampled_data_channel.new_receiver()
            resampled_data_sender = resampled_data_channel.new_sender()

            send_task = asyncio.create_task(send_mock_data(resampled_data_sender))

            async with MovingWindow(
                size=timedelta(seconds=5),
                resampled_data_recv=resampled_data_receiver,
                input_sampling_period=timedelta(seconds=1),
            ) as window:
                time_start = datetime.now(tz=timezone.utc)
                time_end = time_start + timedelta(seconds=5)

                # ... wait for 5 seconds until the buffer is filled
                await asyncio.sleep(5)

                # return an numpy array from the window
                array = window[time_start:time_end]
                # and use it to for example calculate the mean
                mean = array.mean()

        asyncio.run(run())
        ```

    Example: Create a polars data frame from a `MovingWindow`

        ```python
        from datetime import datetime, timedelta, timezone

        async def send_mock_data(sender: Sender[Sample]) -> None:
            while True:
                await sender.send(Sample(datetime.now(tz=timezone.utc), 10.0))
                await asyncio.sleep(1.0)

        async def run() -> None:
            resampled_data_channel = Broadcast[Sample](name="sample-data")
            resampled_data_receiver = resampled_data_channel.new_receiver()
            resampled_data_sender = resampled_data_channel.new_sender()

            send_task = asyncio.create_task(send_mock_data(resampled_data_sender))

            # create a window that stores two days of data
            # starting at 1.1.23 with samplerate=1
            async with MovingWindow(
                size=timedelta(days=2),
                resampled_data_recv=resampled_data_receiver,
                input_sampling_period=timedelta(seconds=1),
            ) as window:
                # wait for one full day until the buffer is filled
                await asyncio.sleep(60*60*24)

                time_start = datetime(2023, 1, 1, tzinfo=timezone.utc)
                time_end = datetime(2023, 1, 2, tzinfo=timezone.utc)

                # You can now create a polars series with one full day of data by
                # passing the window slice, like:
                # series = pl.Series("Jan_1", window[time_start:time_end])

        asyncio.run(run())
        ```
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        size: timedelta,
        resampled_data_recv: Receiver[Sample[Quantity]],
        input_sampling_period: timedelta,
        resampler_config: ResamplerConfig | None = None,
        align_to: datetime = UNIX_EPOCH,
        name: str | None = None,
    ) -> None:
        """
        Initialize the MovingWindow.

        This method creates the underlying ring buffer and starts a
        new task that updates the ring buffer with new incoming samples.
        The task stops running only if the channel receiver is closed.

        Args:
            size: The time span of the moving window over which samples will be stored.
            resampled_data_recv: A receiver that delivers samples with a
                given sampling period.
            input_sampling_period: The time interval between consecutive input samples.
            resampler_config: The resampler configuration in case resampling is required.
            align_to: A timestamp that defines a point in time to which
                the window is aligned to modulo window size. For further
                information, consult the class level documentation.
            name: The name of this moving window. If `None`, `str(id(self))` will be
                used. This is used mostly for debugging purposes.
        """
        assert (
            input_sampling_period.total_seconds() > 0
        ), "The input sampling period should be greater than zero."
        assert (
            input_sampling_period <= size
        ), "The input sampling period should be equal to or lower than the window size."
        super().__init__(name=name)

        self._sampling_period = input_sampling_period

        self._resampler: Resampler | None = None
        self._resampler_sender: Sender[Sample[Quantity]] | None = None

        if resampler_config:
            assert (
                resampler_config.resampling_period <= size
            ), "The resampling period should be equal to or lower than the window size."

            self._resampler = Resampler(resampler_config)
            self._sampling_period = resampler_config.resampling_period

        # Sampling period might not fit perfectly into the window size.
        num_samples = math.ceil(
            size.total_seconds() / self._sampling_period.total_seconds()
        )

        self._resampled_data_recv = resampled_data_recv
        self._buffer = OrderedRingBuffer(
            np.empty(shape=num_samples, dtype=float),
            sampling_period=self._sampling_period,
            align_to=align_to,
        )

    def start(self) -> None:
        """Start the MovingWindow.

        This method starts the MovingWindow tasks.
        """
        if self._resampler:
            self._configure_resampler()
        self._tasks.add(asyncio.create_task(self._run_impl(), name="update-window"))

    @property
    def sampling_period(self) -> timedelta:
        """
        Return the sampling period of the MovingWindow.

        Returns:
            The sampling period of the MovingWindow.
        """
        return self._sampling_period

    @property
    def oldest_timestamp(self) -> datetime | None:
        """
        Return the oldest timestamp of the MovingWindow.

        Returns:
            The oldest timestamp of the MovingWindow or None if the buffer is empty.
        """
        return self._buffer.oldest_timestamp

    @property
    def newest_timestamp(self) -> datetime | None:
        """
        Return the newest timestamp of the MovingWindow.

        Returns:
            The newest timestamp of the MovingWindow or None if the buffer is empty.
        """
        return self._buffer.newest_timestamp

    @property
    def capacity(self) -> int:
        """
        Return the capacity of the MovingWindow.

        Capacity is the maximum number of samples that can be stored in the
        MovingWindow.

        Returns:
            The capacity of the MovingWindow.
        """
        return self._buffer.maxlen

    # pylint before 3.0 only accepts names with 3 or more chars
    def at(self, key: int | datetime) -> float:  # pylint: disable=invalid-name
        """
        Return the sample at the given index or timestamp.

        In contrast to the [`window`][frequenz.sdk.timeseries.MovingWindow.window] method,
        which expects a slice as argument, this method expects a single index as argument
        and returns a single value.

        Args:
            key: The index or timestamp of the sample to return.

        Returns:
            The sample at the given index or timestamp.

        Raises:
            IndexError: If the buffer is empty or the index is out of bounds.
        """
        if self._buffer.count_valid() == 0:
            raise IndexError("The buffer is empty.")

        if isinstance(key, datetime):
            assert self._buffer.oldest_timestamp is not None
            assert self._buffer.newest_timestamp is not None
            if (
                key < self._buffer.oldest_timestamp
                or key > self._buffer.newest_timestamp
            ):
                raise IndexError(
                    f"Timestamp {key} is out of range [{self._buffer.oldest_timestamp}, "
                    f"{self._buffer.newest_timestamp}]"
                )
            return self._buffer[self._buffer.to_internal_index(key)]

        if isinstance(key, int):
            _logger.debug("Returning value at index %s ", key)
            timestamp = self._buffer.get_timestamp(key)
            assert timestamp is not None
            return self._buffer[self._buffer.to_internal_index(timestamp)]

        assert_never(key)

    def window(
        self,
        start: datetime | int | None,
        end: datetime | int | None,
        *,
        force_copy: bool = True,
        fill_value: float | None = np.nan,
    ) -> ArrayLike:
        """
        Return an array containing the samples in the given time interval.

        In contrast to the [`at`][frequenz.sdk.timeseries.MovingWindow.at] method,
        which expects a single index as argument, this method expects a slice as argument
        and returns an array.

        Args:
            start: The start timestamp of the time interval. If `None`, the
                start of the window is used.
            end: The end timestamp of the time interval. If `None`, the end of
                the window is used.
            force_copy: If `True`, the returned array is a copy of the underlying
                data. Otherwise, if possible, a view of the underlying data is
                returned.
            fill_value: If not None, will use this value to fill missing values.
                If missing values should be set, force_copy must be True.
                Defaults to NaN to avoid returning outdated data unexpectedly.

        Returns:
            An array containing the samples in the given time interval.
        """
        return self._buffer.window(
            start, end, force_copy=force_copy, fill_value=fill_value
        )

    async def _run_impl(self) -> None:
        """Awaits samples from the receiver and updates the underlying ring buffer.

        Raises:
            asyncio.CancelledError: if the MovingWindow task is cancelled.
        """
        try:
            async for sample in self._resampled_data_recv:
                _logger.debug("Received new sample: %s", sample)
                if self._resampler and self._resampler_sender:
                    await self._resampler_sender.send(sample)
                else:
                    self._buffer.update(sample)

        except asyncio.CancelledError:
            _logger.info("MovingWindow task has been cancelled.")
            raise

        _logger.error("Channel has been closed")

    def _configure_resampler(self) -> None:
        """Configure the components needed to run the resampler."""
        assert self._resampler is not None

        async def sink_buffer(sample: Sample[Quantity]) -> None:
            if sample.value is not None:
                self._buffer.update(sample)

        resampler_channel = Broadcast[Sample[Quantity]](name="average")
        self._resampler_sender = resampler_channel.new_sender()
        self._resampler.add_timeseries(
            "avg", resampler_channel.new_receiver(), sink_buffer
        )
        self._tasks.add(
            asyncio.create_task(self._resampler.resample(), name="resample")
        )

    def count_valid(self) -> int:
        """
        Count the number of valid samples in this `MovingWindow`.

        Returns:
            The number of valid samples in this `MovingWindow`.
        """
        return self._buffer.count_valid()

    def count_covered(self) -> int:
        """Count the number of samples that are covered by the oldest and newest valid samples.

        Returns:
            The count of samples between the oldest and newest (inclusive) valid samples
                or 0 if there are is no time range covered.
        """
        return self._buffer.count_covered()

    @overload
    def __getitem__(self, key: SupportsIndex) -> float:
        """See the main __getitem__ method."""

    @overload
    def __getitem__(self, key: datetime) -> float:
        """See the main __getitem__ method."""

    @overload
    def __getitem__(self, key: slice) -> ArrayLike:
        """See the main __getitem__ method."""

    # We need the noqa because `IndexError` is raised indirectly by `at()` and `window()`
    def __getitem__(  # noqa: DOC503
        self, key: SupportsIndex | datetime | slice
    ) -> float | ArrayLike:
        """
        Return a sub window of the `MovingWindow`.

        The `MovingWindow` is accessed either by timestamp or by index
        or by a slice of timestamps or integers.

        * If the key is an integer, the float value of that key
          at the given position is returned.
        * If the key is a timestamp, the float value of that key
          that corresponds to the timestamp is returned.
        * If the key is a slice of timestamps or integers, an ndarray is returned,
          where the bounds correspond to the slice bounds.
          Note that a half open interval, which is open at the end, is returned.

        Note:
            Slicing with a step other than 1 is not supported.

        Args:
            key: Either an integer or a timestamp or a slice of timestamps or integers.

        Raises:
            IndexError: when requesting an out of range timestamp or index
            ValueError: when requesting a slice with a step other than 1

        Returns:
            A float if the key is a number or a timestamp.
            an numpy array if the key is a slice.
        """
        if isinstance(key, slice):
            if not (key.step is None or key.step == 1):
                raise ValueError("Slicing with a step other than 1 is not supported.")
            return self.window(key.start, key.stop)

        if isinstance(key, datetime):
            return self.at(key)

        if isinstance(key, SupportsIndex):
            return self.at(key.__index__())

        assert_never(key)


# We need to register the class as a subclass of Sequence like this because
# otherwise type-checking fails complaining that MovingWindow has more
# overloads of __getitem__() than Sequence (which doesn't have an overload with
# a datetime key)
Sequence.register(MovingWindow)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Base class for formula generators that use the component graphs."""

from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from collections import abc
from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic

from frequenz.channels import Sender
from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId

from ...._internal._channels import ChannelRegistry
from ....microgrid import connection_manager
from ....microgrid._data_sourcing import ComponentMetricRequest
from ..._base_types import QuantityT
from .._formula_engine import FormulaEngine, FormulaEngine3Phase
from .._resampled_formula_builder import ResampledFormulaBuilder


class FormulaGenerationError(Exception):
    """An error encountered during formula generation from the component graph."""


class ComponentNotFound(FormulaGenerationError):
    """Indicates that a component required for generating a formula is not found."""


NON_EXISTING_COMPONENT_ID = sys.maxsize
"""The component ID for non-existent components in the components graph.

The non-existing component ID is commonly used in scenarios where a formula
engine requires a component ID but there are no available components in the
graph to associate with it. Thus, the non-existing component ID is subscribed
instead so that the formula engine can send `None` or `0` values at the same
frequency as the other streams.
"""


@dataclass(frozen=True)
class FormulaGeneratorConfig:
    """Config for formula generators."""

    component_ids: abc.Set[int] | None = None
    """The component IDs to use for generating the formula."""

    allow_fallback: bool = True


class FormulaGenerator(ABC, Generic[QuantityT]):
    """A class for generating formulas from the component graph."""

    def __init__(
        self,
        namespace: str,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
        config: FormulaGeneratorConfig,
    ) -> None:
        """Create a `FormulaGenerator` instance.

        Args:
            namespace: A namespace to use with the data-pipeline.
            channel_registry: A channel registry instance shared with the resampling
                actor.
            resampler_subscription_sender: A sender for sending metric requests to the
                resampling actor.
            config: configs for the formula generator.
        """
        self._channel_registry: ChannelRegistry = channel_registry
        self._resampler_subscription_sender: Sender[ComponentMetricRequest] = (
            resampler_subscription_sender
        )
        self._namespace: str = namespace
        self._config: FormulaGeneratorConfig = config

    @property
    def namespace(self) -> str:
        """Get the namespace for the formula generator."""
        return self._namespace

    def _get_builder(
        self,
        name: str,
        component_metric_id: ComponentMetricId,
        create_method: Callable[[float], QuantityT],
    ) -> ResampledFormulaBuilder[QuantityT]:
        builder = ResampledFormulaBuilder(
            namespace=self._namespace,
            formula_name=name,
            channel_registry=self._channel_registry,
            resampler_subscription_sender=self._resampler_subscription_sender,
            metric_id=component_metric_id,
            create_method=create_method,
        )
        return builder

    def _get_grid_component(self) -> Component:
        """
        Get the grid component in the component graph.

        Returns:
            The first grid component found in the graph.

        Raises:
            ComponentNotFound: If the grid component is not found in the component graph.
        """
        component_graph = connection_manager.get().component_graph
        grid_component = next(
            iter(
                component_graph.components(
                    component_categories={ComponentCategory.GRID}
                )
            ),
            None,
        )
        if grid_component is None:
            raise ComponentNotFound("Grid component not found in the component graph.")

        return grid_component

    def _get_grid_component_successors(self) -> set[Component]:
        """Get the set of grid component successors in the component graph.

        Returns:
            A set of grid component successors.

        Raises:
            ComponentNotFound: If no successor components are found in the component graph.
        """
        grid_component = self._get_grid_component()
        component_graph = connection_manager.get().component_graph
        grid_successors = component_graph.successors(grid_component.component_id)

        if not grid_successors:
            raise ComponentNotFound("No components found in the component graph.")

        return grid_successors

    @abstractmethod
    def generate(
        self,
    ) -> FormulaEngine[QuantityT] | FormulaEngine3Phase[QuantityT]:
        """Generate a formula engine, based on the component graph."""

    def _get_metric_fallback_components(
        self, components: set[Component]
    ) -> dict[Component, set[Component]]:
        """Get primary and fallback components within a given set of components.

        When a meter is positioned before one or more components of the same type (e.g., inverters),
        it is considered the primary component, and the components that follow are treated
        as fallback components.
        If the non-meter component has no meter in front of it, then it is the primary component
        and has no fallbacks.

        The method iterates through the provided components and assesses their roles as primary
        or fallback components.
        If a component:
         * can act as a primary component (e.g., a meter), then it finds its
        fallback components and pairs them together.
         * can act as a fallback (e.g., an inverter or EV charger), then it finds
        the primary component for it (usually a meter) and pairs them together.
         * has no fallback (e.g., an inverter that has no meter attached), then it
        returns an empty set for that component. This means that the component
        is a primary component and has no fallbacks.

        Args:
            components: The components to be analyzed.

        Returns:
            A dictionary where:
                * The keys are primary components.
                * The values are sets of fallback components.
        """
        graph = connection_manager.get().component_graph
        fallbacks: dict[Component, set[Component]] = {}

        for component in components:
            if component.category == ComponentCategory.METER:
                fallbacks[component] = self._get_meter_fallback_components(component)
            else:
                predecessors = graph.predecessors(component.component_id)
                if len(predecessors) == 1:
                    predecessor = predecessors.pop()
                    if self._is_primary_fallback_pair(predecessor, component):
                        # predecessor is primary component and the component is one of the
                        # fallbacks components.
                        fallbacks.setdefault(predecessor, set()).add(component)
                        continue

                # This component is primary component with no fallbacks.
                fallbacks[component] = set()
        return fallbacks

    def _get_meter_fallback_components(self, meter: Component) -> set[Component]:
        """Get the fallback components for a given meter.

        Args:
            meter: The meter to find the fallback components for.

        Returns:
            A set of fallback components for the given meter.
            An empty set is returned if the meter has no fallbacks.
        """
        assert meter.category == ComponentCategory.METER

        graph = connection_manager.get().component_graph
        successors = graph.successors(meter.component_id)

        # All fallbacks has to be of the same type and category.
        if (
            all(graph.is_chp(c) for c in successors)
            or all(graph.is_pv_inverter(c) for c in successors)
            or all(graph.is_battery_inverter(c) for c in successors)
            or all(graph.is_ev_charger(c) for c in successors)
        ):
            return successors
        return set()

    def _is_primary_fallback_pair(
        self,
        primary_candidate: Component,
        fallback_candidate: Component,
    ) -> bool:
        """Determine if a given component can act as a primary-fallback pair.

        This method checks:
         * whether the `fallback_candidate` is of a type that can have the `primary_candidate`,
         * if `primary_candidate` is the primary measuring point of the `fallback_candidate`.

        Args:
            primary_candidate: The component to be checked as a primary measuring device.
            fallback_candidate: The component to be checked as a fallback measuring device.

        Returns:
            bool: True if the provided components are a primary-fallback pair, False otherwise.
        """
        graph = connection_manager.get().component_graph

        # reassign to decrease the length of the line and make code readable
        fallback = fallback_candidate
        primary = primary_candidate

        # fmt: off
        return (
            graph.is_pv_inverter(fallback) and graph.is_pv_meter(primary)
            or graph.is_chp(fallback) and graph.is_chp_meter(primary)
            or graph.is_ev_charger(fallback) and graph.is_ev_charger_meter(primary)
            or graph.is_battery_inverter(fallback) and graph.is_battery_meter(primary)
        )
        # fmt: on



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""General purpose classes for use with channels."""

import abc
import dataclasses
import logging
import traceback
import typing

from frequenz.channels import Broadcast, Receiver

_logger = logging.getLogger(__name__)

T_co = typing.TypeVar("T_co", covariant=True)
U_co = typing.TypeVar("U_co", covariant=True)
T = typing.TypeVar("T")


class ReceiverFetcher(typing.Generic[T_co], typing.Protocol):
    """An interface that just exposes a `new_receiver` method."""

    @abc.abstractmethod
    def new_receiver(self, *, limit: int = 50) -> Receiver[T_co]:
        """Get a receiver from the channel.

        Args:
            limit: The maximum size of the receiver.

        Returns:
            A receiver instance.
        """


class MappingReceiverFetcher(typing.Generic[T_co, U_co]):
    """A receiver fetcher that can manipulate receivers before returning them."""

    def __init__(
        self,
        fetcher: ReceiverFetcher[T_co],
        mapping_function: typing.Callable[[Receiver[T_co]], Receiver[U_co]],
    ) -> None:
        """Initialize this instance.

        Args:
            fetcher: The underlying fetcher to get receivers from.
            mapping_function: The method to be applied on new receivers before returning
                them.
        """
        self._fetcher = fetcher
        self._mapping_function = mapping_function

    def new_receiver(self, *, limit: int = 50) -> Receiver[U_co]:
        """Get a receiver from the channel.

        Args:
            limit: The maximum size of the receiver.

        Returns:
            A receiver instance.
        """
        return self._mapping_function(self._fetcher.new_receiver(limit=limit))


class ChannelRegistry:
    """Dynamically creates, own and provide access to broadcast channels.

    It can be used by actors to dynamically establish a communication channel
    between each other.

    The registry is responsible for creating channels when they are first requested via
    the [`get_or_create()`][frequenz.sdk.actor.ChannelRegistry.get_or_create] method.

    The registry also stores type information to make sure that the same channel is not
    used for different message types.

    Since the registry owns the channels, it is also responsible for closing them when
    they are no longer needed. There is no way to remove a channel without closing it.

    Note:
        This registry stores [`Broadcast`][frequenz.channels.Broadcast] channels.
    """

    def __init__(self, *, name: str) -> None:
        """Initialize this registry.

        Args:
            name: A name to identify the registry in the logs. This name is also used as
                a prefix for the channel names.
        """
        self._name = name
        self._channels: dict[str, _Entry] = {}

    @property
    def name(self) -> str:
        """The name of this registry."""
        return self._name

    def message_type(self, key: str) -> type:
        """Get the message type of the channel for the given key.

        Args:
            key: The key to identify the channel.

        Returns:
            The message type of the channel.

        Raises:
            KeyError: If the channel does not exist.
        """
        entry = self._channels.get(key)
        if entry is None:
            raise KeyError(f"No channel for key {key!r} exists.")
        return entry.message_type

    def __contains__(self, key: str) -> bool:
        """Check whether the channel for the given `key` exists."""
        return key in self._channels

    def get_or_create(self, message_type: type[T], key: str) -> Broadcast[T]:
        """Get or create a channel for the given key.

        If a channel for the given key already exists, the message type of the existing
        channel is checked against the requested message type. If they do not match,
        a `ValueError` is raised.

        Note:
            The types have to match exactly, it doesn't do a subtype check due to
            technical limitations. In the future subtype checks might be supported.

        Args:
            message_type: The type of the message that is sent through the channel.
            key: The key to identify the channel.

        Returns:
            The channel for the given key.

        Raises:
            ValueError: If the channel exists and the message type does not match.
        """
        if key not in self._channels:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    "Creating a new channel for key %r with type %s at:\n%s",
                    key,
                    message_type,
                    "".join(traceback.format_stack(limit=10)[:9]),
                )
            self._channels[key] = _Entry(
                message_type, Broadcast(name=f"{self._name}-{key}")
            )

        entry = self._channels[key]
        if entry.message_type is not message_type:
            error_message = (
                f"Type mismatch, a channel for key {key!r} exists and the requested "
                f"message type {message_type} is not the same as the existing "
                f"message type {entry.message_type}."
            )
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    "%s at:\n%s",
                    error_message,
                    # We skip the last frame because it's this method, and limit the
                    # stack to 9 frames to avoid adding too much noise.
                    "".join(traceback.format_stack(limit=10)[:9]),
                )
            raise ValueError(error_message)

        return typing.cast(Broadcast[T], entry.channel)

    async def close_and_remove(self, key: str) -> None:
        """Remove the channel for the given key.

        Args:
            key: The key to identify the channel.

        Raises:
            KeyError: If the channel does not exist.
        """
        entry = self._channels.pop(key, None)
        if entry is None:
            raise KeyError(f"No channel for key {key!r} exists.")
        await entry.channel.close()


@dataclasses.dataclass(frozen=True)
class _Entry:
    """An entry in a channel registry."""

    message_type: type
    """The type of the message that is sent through the channel in this entry."""

    # We use object instead of Any to minimize the chances of hindering type checking.
    # If for some reason the channel is not casted to the proper underlaying type, when
    # using object at least accessing any member that's not part of the object base
    # class will yield a type error, while if we used Any, it would not and the issue
    # would be much harder to find.
    channel: Broadcast[object]
    """The channel in this entry."""



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""A tokenizer for data pipeline formulas."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


class StringIter:
    """An iterator for reading characters from a string."""

    def __init__(self, raw: str) -> None:
        """Create a `StringIter` instance.

        Args:
            raw: The raw string to create the iterator out of.
        """
        self._raw = raw
        self._max = len(raw)
        self._pos = 0

    @property
    def pos(self) -> int:
        """Return the position of the iterator in the raw string.

        Returns:
            The position of the iterator.
        """
        return self._pos

    @property
    def raw(self) -> str:
        """Return the raw string the iterator is created with.

        Returns:
            The base string of the iterator.
        """
        return self._raw

    def __iter__(self) -> StringIter:
        """Return an iterator to this class.

        Returns:
            self.
        """
        return self

    def __next__(self) -> str:
        """Return the next character in the raw string, and move forward.

        Returns:
            The next character.

        Raises:
            StopIteration: when there are no more characters in the string.
        """
        if self._pos < self._max:
            char = self._raw[self._pos]
            self._pos += 1
            return char
        raise StopIteration()

    def peek(self) -> str | None:
        """Return the next character in the raw string, without consuming it.

        Returns:
            The next character.
        """
        if self._pos < self._max:
            return self._raw[self._pos]
        return None


class TokenType(Enum):
    """Represents the types of tokens the Tokenizer can return."""

    COMPONENT_METRIC = 0
    """A component metric ID."""

    CONSTANT = 1
    """A constant value."""

    OPER = 2
    """An operator."""


@dataclass
class Token:
    """Represents a Token returned by the Tokenizer."""

    type: TokenType
    """The type of the token."""

    value: str
    """The value associated to the token."""


class Tokenizer:
    """A Tokenizer for breaking down a string formula into individual tokens.

    Every instance is an iterator that allows us to iterate over the individual tokens
    in the given formula.

    Formulas can have Component IDs that are preceeded by a pound symbol("#"), and these
    operators: +, -, *, /, (, ).

    For example, the input string: "#20 + #5" would produce three tokens:
     - COMPONENT_METRIC: 20
     - OPER: +
     - COMPONENT_METRIC: 5
    """

    def __init__(self, formula: str) -> None:
        """Create a `Tokenizer` instance.

        Args:
            formula: The string formula to tokenize.
        """
        self._formula = StringIter(formula)

    def _read_unsigned_int(self) -> str:
        """Read an unsigned int from the current position in the input string.

        Returns:
            A string containing the read unsigned int value.

        Raises:
            ValueError: when there is no unsigned int at the current position.
        """
        first_char = True
        result = ""

        while char := self._formula.peek():
            if not char.isdigit():
                if first_char:
                    raise ValueError(
                        f"Expected an integer. got '{char}', "
                        f"at pos {self._formula.pos} in formula {self._formula.raw}"
                    )
                break
            first_char = False
            result += char
            next(self._formula)
        return result

    def __iter__(self) -> Tokenizer:
        """Return an iterator to this class.

        Returns:
            self.
        """
        return self

    def __next__(self) -> Token:
        """Return the next token in the input string.

        Returns:
            The next token.

        Raises:
            ValueError: when there are unknown tokens in the input string.
            StopIteration: when there are no more tokens in the input string.
        """
        for char in self._formula:
            if char in (" ", "\n", "\r", "\t"):
                continue
            if char in ("+", "-", "*", "/", "(", ")"):
                return Token(TokenType.OPER, char)
            if char == "#":
                return Token(TokenType.COMPONENT_METRIC, self._read_unsigned_int())
            raise ValueError(
                f"Unable to parse character '{char}' at pos: {self._formula.pos}"
                f" in formula: {self._formula.raw}"
            )
        raise StopIteration()



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Manage PV inverters for the power distributor."""

import asyncio
import collections.abc
import logging
from datetime import timedelta

from frequenz.channels import Broadcast, LatestValueCache, Sender
from frequenz.client.microgrid import (
    ApiClientError,
    ComponentCategory,
    InverterData,
    InverterType,
)
from frequenz.quantities import Power
from typing_extensions import override

from ....._internal._math import is_close_to_zero
from .... import connection_manager
from ..._component_pool_status_tracker import ComponentPoolStatusTracker
from ..._component_status import ComponentPoolStatus, PVInverterStatusTracker
from ...request import Request
from ...result import PartialFailure, Result, Success
from .._component_manager import ComponentManager

_logger = logging.getLogger(__name__)


class PVManager(ComponentManager):
    """Manage PV inverters for the power distributor."""

    @override
    def __init__(
        self,
        component_pool_status_sender: Sender[ComponentPoolStatus],
        results_sender: Sender[Result],
        api_power_request_timeout: timedelta,
    ) -> None:
        """Initialize this instance.

        Args:
            component_pool_status_sender: Channel for sending information about which
                components are expected to be working.
            results_sender: Channel for sending results of power distribution.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
        """
        self._results_sender = results_sender
        self._api_power_request_timeout = api_power_request_timeout
        self._pv_inverter_ids = self._get_pv_inverter_ids()

        self._component_pool_status_tracker = (
            ComponentPoolStatusTracker(
                component_ids=self._pv_inverter_ids,
                component_status_sender=component_pool_status_sender,
                max_data_age=timedelta(seconds=10.0),
                max_blocking_duration=timedelta(seconds=30.0),
                component_status_tracker_type=PVInverterStatusTracker,
            )
            if self._pv_inverter_ids
            else None
        )
        self._component_data_caches: dict[int, LatestValueCache[InverterData]] = {}
        self._target_power = Power.zero()
        self._target_power_channel = Broadcast[Request](name="target_power")
        self._target_power_tx = self._target_power_channel.new_sender()
        self._task: asyncio.Task[None] | None = None

    @override
    def component_ids(self) -> collections.abc.Set[int]:
        """Return the set of PV inverter ids."""
        return self._pv_inverter_ids

    @override
    async def start(self) -> None:
        """Start the PV inverter manager."""
        self._component_data_caches = {
            inv_id: LatestValueCache(
                await connection_manager.get().api_client.inverter_data(inv_id),
                unique_id=f"{type(self).__name__}«{hex(id(self))}»:inverter«{inv_id}»",
            )
            for inv_id in self._pv_inverter_ids
        }

    @override
    async def stop(self) -> None:
        """Stop the PV inverter manager."""
        await asyncio.gather(
            *[tracker.stop() for tracker in self._component_data_caches.values()]
        )
        if self._component_pool_status_tracker:
            await self._component_pool_status_tracker.stop()

    @override
    async def distribute_power(self, request: Request) -> None:
        """Distribute the requested power to the PV inverters.

        Args:
            request: Request to get the distribution for.

        Raises:
            ValueError: If no PV inverters are present in the component graph, but
                component_ids are provided in the request.
        """
        remaining_power = request.power
        allocations: dict[int, Power] = {}
        if not self._component_pool_status_tracker:
            if not request.component_ids:
                await self._results_sender.send(
                    Success(
                        succeeded_components=set(),
                        succeeded_power=Power.zero(),
                        excess_power=remaining_power,
                        request=request,
                    )
                )
                return
            raise ValueError(
                "Cannot distribute power to PV inverters without any inverters"
            )

        working_components: list[int] = []
        for inv_id in self._component_pool_status_tracker.get_working_components(
            request.component_ids
        ):
            if self._component_data_caches[inv_id].has_value():
                working_components.append(inv_id)
            else:
                _logger.warning(
                    "Excluding PV inverter %s from power distribution due to "
                    "lack of data since startup.",
                    inv_id,
                )

        # When sorting by lower bounds, which are negative for PV inverters, we have to
        # reverse the order, so that the inverters with the higher bounds i.e., the
        # least absolute value are first.
        working_components.sort(
            key=lambda inv_id: self._component_data_caches[inv_id]
            .get()
            .active_power_inclusion_lower_bound,
            reverse=True,
        )

        num_components = len(working_components)
        if num_components == 0:
            _logger.error("No PV inverters available for power distribution. Aborting.")
            return

        for idx, inv_id in enumerate(working_components):
            # Request powers are negative for PV inverters.  When remaining power is
            # greater than or equal to 0.0, we can stop allocating further, and set 0
            # power for all inverters for which no allocations were made.
            if remaining_power > Power.zero() or is_close_to_zero(
                remaining_power.as_watts()
            ):
                allocations[inv_id] = Power.zero()
                continue
            distribution = remaining_power / float(num_components - idx)
            inv_data = self._component_data_caches[inv_id]
            if not inv_data.has_value():
                allocations[inv_id] = Power.zero()
                # Can't get device bounds, so can't use inverter.
                continue
            discharge_bounds = Power.from_watts(
                inv_data.get().active_power_inclusion_lower_bound
            )
            # Because all 3 values are negative or 0, we use max, to get the value
            # with the least absolute value.
            allocated_power = max(remaining_power, discharge_bounds, distribution)
            allocations[inv_id] = allocated_power
            remaining_power -= allocated_power

        _logger.debug(
            "Distributing %s to PV inverters %s",
            request.power,
            allocations,
        )
        await self._set_api_power(request, allocations, remaining_power)

    async def _set_api_power(  # pylint: disable=too-many-locals
        self, request: Request, allocations: dict[int, Power], remaining_power: Power
    ) -> None:
        api_client = connection_manager.get().api_client
        tasks: dict[int, asyncio.Task[None]] = {}
        for component_id, power in allocations.items():
            tasks[component_id] = asyncio.create_task(
                api_client.set_power(component_id, power.as_watts())
            )
        _, pending = await asyncio.wait(
            tasks.values(),
            timeout=self._api_power_request_timeout.total_seconds(),
            return_when=asyncio.ALL_COMPLETED,
        )
        # collect the timed out tasks and cancel them while keeping the
        # exceptions, so that they can be processed later.
        for task in pending:
            task.cancel()
        await asyncio.gather(*pending, return_exceptions=True)

        failed_components: set[int] = set()
        succeeded_components: set[int] = set()
        failed_power = Power.zero()
        for component_id, task in tasks.items():
            try:
                task.result()
            except asyncio.CancelledError:
                _logger.warning(
                    "Timeout while setting power to PV inverter %s", component_id
                )
            except ApiClientError as exc:
                _logger.warning(
                    "Got a client error while setting power to PV inverter %s: %s",
                    component_id,
                    exc,
                )
            except Exception:  # pylint: disable=broad-except
                _logger.exception(
                    "Unknown error while setting power to PV inverter: %s",
                    component_id,
                )
            else:
                succeeded_components.add(component_id)
                continue

            failed_components.add(component_id)
            failed_power += allocations[component_id]

        if failed_components:
            await self._results_sender.send(
                PartialFailure(
                    failed_components=failed_components,
                    succeeded_components=succeeded_components,
                    failed_power=failed_power,
                    succeeded_power=self._target_power - failed_power,
                    excess_power=remaining_power,
                    request=request,
                )
            )
            return
        await self._results_sender.send(
            Success(
                succeeded_components=succeeded_components,
                succeeded_power=self._target_power,
                excess_power=remaining_power,
                request=request,
            )
        )

    def _get_pv_inverter_ids(self) -> collections.abc.Set[int]:
        """Return the IDs of all PV inverters present in the component graph."""
        return {
            inv.component_id
            for inv in connection_manager.get().component_graph.components(
                component_categories={ComponentCategory.INVERTER}
            )
            if inv.type == InverterType.SOLAR
        }



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Generators for formulas from component graphs."""

from ._battery_power_formula import BatteryPowerFormula
from ._chp_power_formula import CHPPowerFormula
from ._consumer_power_formula import ConsumerPowerFormula
from ._ev_charger_current_formula import EVChargerCurrentFormula
from ._ev_charger_power_formula import EVChargerPowerFormula
from ._formula_generator import (
    ComponentNotFound,
    FormulaGenerationError,
    FormulaGenerator,
    FormulaGeneratorConfig,
)
from ._grid_current_formula import GridCurrentFormula
from ._grid_power_3_phase_formula import GridPower3PhaseFormula
from ._grid_power_formula import GridPowerFormula
from ._grid_reactive_power_formula import GridReactivePowerFormula
from ._producer_power_formula import ProducerPowerFormula
from ._pv_power_formula import PVPowerFormula

__all__ = [
    #
    # Base class
    #
    "FormulaGenerator",
    "FormulaGeneratorConfig",
    #
    # Power Formula generators
    #
    "CHPPowerFormula",
    "ConsumerPowerFormula",
    "GridPower3PhaseFormula",
    "GridPowerFormula",
    "GridReactivePowerFormula",
    "BatteryPowerFormula",
    "EVChargerPowerFormula",
    "PVPowerFormula",
    "ProducerPowerFormula",
    #
    # Current formula generators
    #
    "GridCurrentFormula",
    "EVChargerCurrentFormula",
    #
    # Exceptions
    #
    "ComponentNotFound",
    "FormulaGenerationError",
]



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Manages shared state/tasks for a set of EV chargers."""

import asyncio
import uuid
from collections import abc

from frequenz.channels import Broadcast, Receiver, Sender
from frequenz.client.microgrid import ComponentCategory

from ..._internal._channels import ChannelRegistry, ReceiverFetcher
from ...microgrid import connection_manager
from ...microgrid._data_sourcing import ComponentMetricRequest
from ...microgrid._power_distributing import ComponentPoolStatus, Result
from ...microgrid._power_managing._base_classes import Proposal, ReportRequest
from .._base_types import SystemBounds
from ..formula_engine._formula_engine_pool import FormulaEnginePool
from ._system_bounds_tracker import EVCSystemBoundsTracker


class EVChargerPoolReferenceStore:
    """A class for maintaining the shared state/tasks for a set of pool of EV chargers.

    This includes ownership of
    - the formula engine pool and metric calculators.
    - the tasks for calculating system bounds for the EV chargers.

    These are independent of the priority of the actors and can be shared between
    multiple users of the same set of EV chargers.

    They are exposed through the EVChargerPool class.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
        status_receiver: Receiver[ComponentPoolStatus],
        power_manager_requests_sender: Sender[Proposal],
        power_manager_bounds_subs_sender: Sender[ReportRequest],
        power_distribution_results_fetcher: ReceiverFetcher[Result],
        component_ids: abc.Set[int] | None = None,
    ):
        """Create an instance of the class.

        Args:
            channel_registry: A channel registry instance shared with the resampling
                actor.
            resampler_subscription_sender: A sender for sending metric requests to the
                resampling actor.
            status_receiver: A receiver that streams the status of the EV Chargers in
                the pool.
            power_manager_requests_sender: A Channel sender for sending power
                requests to the power managing actor.
            power_manager_bounds_subs_sender: A Channel sender for sending power bounds
                subscription requests to the power managing actor.
            power_distribution_results_fetcher: A ReceiverFetcher for the results from
                the power distributing actor.
            component_ids: An optional list of component_ids belonging to this pool.  If
                not specified, IDs of all EV Chargers in the microgrid will be fetched
                from the component graph.
        """
        self.channel_registry = channel_registry
        self.resampler_subscription_sender = resampler_subscription_sender
        self.status_receiver = status_receiver
        self.power_manager_requests_sender = power_manager_requests_sender
        self.power_manager_bounds_subs_sender = power_manager_bounds_subs_sender
        self.power_distribution_results_fetcher = power_distribution_results_fetcher

        if component_ids is not None:
            self.component_ids: frozenset[int] = frozenset(component_ids)
        else:
            graph = connection_manager.get().component_graph
            self.component_ids = frozenset(
                {
                    evc.component_id
                    for evc in graph.components(
                        component_categories={ComponentCategory.EV_CHARGER}
                    )
                }
            )

        self.power_bounds_subs: dict[str, asyncio.Task[None]] = {}

        self.namespace: str = f"ev-charger-pool-{uuid.uuid4()}"
        self.formula_pool = FormulaEnginePool(
            self.namespace,
            self.channel_registry,
            self.resampler_subscription_sender,
        )

        self.bounds_channel: Broadcast[SystemBounds] = Broadcast(
            name=f"System Bounds for EV Chargers: {component_ids}",
            resend_latest=True,
        )
        self.bounds_tracker: EVCSystemBoundsTracker = EVCSystemBoundsTracker(
            self.component_ids,
            self.status_receiver,
            self.bounds_channel.new_sender(),
        )
        self.bounds_tracker.start()

    async def stop(self) -> None:
        """Stop all tasks and channels owned by the EVChargerPool."""
        await self.formula_pool.stop()
        await self.bounds_tracker.stop()



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Actor to distribute power between components.

The purpose of this actor is to distribute power between components in a microgrid.

The actor receives power requests from the power manager, process them by
distributing the power between the components and sends the results back to it.
"""


import asyncio
import logging
from datetime import timedelta

from frequenz.channels import Receiver, Sender
from frequenz.client.microgrid import ComponentCategory, ComponentType, InverterType
from typing_extensions import override

from ...actor._actor import Actor
from ._component_managers import (
    BatteryManager,
    ComponentManager,
    EVChargerManager,
    PVManager,
)
from ._component_status import ComponentPoolStatus
from .request import Request
from .result import Result

_logger = logging.getLogger(__name__)


class PowerDistributingActor(Actor):
    # pylint: disable=too-many-instance-attributes
    """Actor to distribute the power between components in a microgrid.

    One instance of the actor can handle only one component category and type,
    which needs to be specified at actor startup and it will setup the correct
    component manager based on the given category and type.

    Only one power request is processed at a time to prevent from sending
    multiple requests for the same components to the microgrid API at the
    same time.

    Edge cases:
    * If a new power request is received while a power request with the same
    set of components is being processed, the new request will be added to
    the pending requests. Then the pending request will be processed after the
    request with the same set of components being processed is done. Only one
    pending request is kept for each set of components, the latest request will
    overwrite the previous one if there is any.

    * If there are 2 requests and their set of components is different but they
    overlap (they have at least one common component), then both requests will
    be processed concurrently. Though, the power manager will make sure this
    doesn't happen as overlapping component IDs are not possible at the moment.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        requests_receiver: Receiver[Request],
        results_sender: Sender[Result],
        component_pool_status_sender: Sender[ComponentPoolStatus],
        *,
        api_power_request_timeout: timedelta,
        component_category: ComponentCategory,
        component_type: ComponentType | None = None,
        name: str | None = None,
    ) -> None:
        """Create actor instance.

        Args:
            requests_receiver: Receiver for receiving power requests from the power
                manager.
            results_sender: Sender for sending results to the power manager.
            component_pool_status_sender: Channel for sending information about which
                components are expected to be working.
            api_power_request_timeout: Timeout to use when making power requests to
                the microgrid API.
            component_category: The category of the components that this actor is
                responsible for.
            component_type: The type of the component of the given category that this
                actor is responsible for.  This is used only when the component category
                is not enough to uniquely identify the component.  For example, when the
                category is `ComponentCategory.INVERTER`, the type is needed to identify
                the inverter as a solar inverter or a battery inverter.  This can be
                `None` when the component category is enough to uniquely identify the
                component.
            name: The name of the actor. If `None`, `str(id(self))` will be used. This
                is used mostly for debugging purposes.

        Raises:
            ValueError: If the given component category is not supported.
        """
        super().__init__(name=name)
        self._component_category = component_category
        self._component_type = component_type
        self._requests_receiver = requests_receiver
        self._result_sender = results_sender
        self._api_power_request_timeout = api_power_request_timeout

        self._processing_tasks: dict[frozenset[int], asyncio.Task[None]] = {}
        """Track the power request tasks currently being processed."""

        self._pending_requests: dict[frozenset[int], Request] = {}
        """Track the power requests that are waiting to be processed.

        Only one pending power request is kept for each set of components, the
        latest request will overwrite the previous one.
        """

        self._component_manager: ComponentManager
        if component_category == ComponentCategory.BATTERY:
            self._component_manager = BatteryManager(
                component_pool_status_sender, results_sender, api_power_request_timeout
            )
        elif component_category == ComponentCategory.EV_CHARGER:
            self._component_manager = EVChargerManager(
                component_pool_status_sender, results_sender, api_power_request_timeout
            )
        elif (
            component_category == ComponentCategory.INVERTER
            and component_type == InverterType.SOLAR
        ):
            self._component_manager = PVManager(
                component_pool_status_sender, results_sender, api_power_request_timeout
            )
        else:
            raise ValueError(
                f"PowerDistributor doesn't support controlling: {component_category}"
            )

    @override
    async def _run(self) -> None:
        """Run this actor's logic.

        It waits for new power requests and process them. Only one power request
        can be processed at a time to prevent from sending multiple requests for
        the same components to the microgrid API at the same time.

        A new power request will be ignored if a power request with the same
        components is currently being processed.

        Every component that failed or didn't respond in time will be marked
        as broken for some time.
        """
        await self._component_manager.start()

        async for request in self._requests_receiver:
            req_id = frozenset(request.component_ids)

            if req_id in self._processing_tasks:
                if pending_request := self._pending_requests.get(req_id):
                    _logger.debug(
                        "Pending request: %s, overwritten with request: %s",
                        pending_request,
                        request,
                    )
                self._pending_requests[req_id] = request
            else:
                self._process_request(req_id, request)

    @override
    async def stop(self, msg: str | None = None) -> None:
        """Stop this actor.

        Args:
            msg: The message to be passed to the tasks being cancelled.
        """
        await self._component_manager.stop()
        await super().stop(msg)

    def _handle_task_completion(
        self, req_id: frozenset[int], request: Request, task: asyncio.Task[None]
    ) -> None:
        """Handle the completion of a power request task.

        Args:
            req_id: The id to identify the power request.
            request: The power request that has been processed.
            task: The task that has completed.
        """
        try:
            task.result()
        except Exception:  # pylint: disable=broad-except
            _logger.exception("Failed power request: %s", request)

        if req_id in self._pending_requests:
            self._process_request(req_id, self._pending_requests.pop(req_id))
        elif req_id in self._processing_tasks:
            del self._processing_tasks[req_id]
        else:
            _logger.error("Request id not found in processing tasks: %s", req_id)

    def _process_request(self, req_id: frozenset[int], request: Request) -> None:
        """Process a power request.

        Args:
            req_id: The id to identify the power request.
            request: The power request to process.
        """
        task = asyncio.create_task(
            self._component_manager.distribute_power(request),
            name=f"{type(self).__name__}:{request}",
        )
        task.add_done_callback(
            lambda t: self._handle_task_completion(req_id, request, t)
        )
        self._processing_tasks[req_id] = task



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""The DataSourcingActor."""

from ._component_metric_request import ComponentMetricId, ComponentMetricRequest
from .data_sourcing import DataSourcingActor

__all__ = [
    "ComponentMetricId",
    "ComponentMetricRequest",
    "DataSourcingActor",
]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula Engine Exceptions."""


class FormulaEngineError(Exception):
    """An error occurred while fetching metrics or applying the formula on them."""



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""General purpose async tools."""


import asyncio
import logging
from abc import ABC
from datetime import timedelta
from typing import Any, Callable, Coroutine

_logger = logging.getLogger(__name__)


async def cancel_and_await(task: asyncio.Task[Any]) -> None:
    """Cancel a task and wait for it to finish.

    Exits immediately if the task is already done.

    The `CancelledError` is suppressed, but any other exception will be propagated.

    Args:
        task: The task to be cancelled and waited for.
    """
    if task.done():
        return
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


async def run_forever(
    async_callable: Callable[[], Coroutine[Any, Any, None]],
    interval: timedelta = timedelta(seconds=1),
) -> None:
    """Run a given function forever, restarting it after any exception.

    Args:
        async_callable: The async callable to run.
        interval: The interval between restarts.
    """
    interval_s = interval.total_seconds()
    while True:
        try:
            await async_callable()
        except Exception:  # pylint: disable=broad-except
            _logger.exception("Restarting after exception")
            await asyncio.sleep(interval_s)


class NotSyncConstructible(AssertionError):
    """Raised when object with async constructor is created in sync way."""


class AsyncConstructible(ABC):
    """Parent class for classes where part of the constructor is async."""

    def __init__(self) -> None:
        """Raise error when object is created in sync way.

        Raises:
            NotSyncConstructible: If this method is called.
        """
        raise NotSyncConstructible(
            "This object shouldn't be created with default constructor. ",
            "Check class documentation for more information.",
        )



# License: MIT
# Copyright © 2024 Frequenz Energy-as-a-Service GmbH

"""Types for exposing PV pool reports."""

import typing

from frequenz.quantities import Power

from .._base_types import Bounds


class PVPoolReport(typing.Protocol):
    """A status report for a PV pool."""

    @property
    def target_power(self) -> Power | None:
        """The currently set power for the PV inverters."""

    @property
    def bounds(self) -> Bounds[Power] | None:
        """The usable bounds for the PV inverters.

        These bounds are adjusted to any restrictions placed by actors with higher
        priorities.
        """



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Ringbuffer implementation & utilities."""

from .buffer import Gap, OrderedRingBuffer
from .serialization import dump, load

__all__ = ["OrderedRingBuffer", "Gap", "load", "dump"]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator for PV Power, from the component graph."""

import logging

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from ....microgrid import connection_manager
from .._formula_engine import FormulaEngine
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import (
    NON_EXISTING_COMPONENT_ID,
    FormulaGenerator,
    FormulaGeneratorConfig,
)

_logger = logging.getLogger(__name__)


class PVPowerFormula(FormulaGenerator[Power]):
    """Creates a formula engine for calculating the PV power production."""

    def generate(  # noqa: DOC502
        # * ComponentNotFound is raised indirectly by _get_pv_power_components
        # * RuntimeError is also raised indirectly by _get_pv_power_components
        self,
    ) -> FormulaEngine[Power]:
        """Make a formula for the PV power production of a microgrid.

        Returns:
            A formula engine that will calculate PV power production values.

        Raises:
            ComponentNotFound: if there is a problem finding the needed components.
            RuntimeError: if the grid component has no PV inverters or meters as
                successors.
        """
        builder = self._get_builder(
            "pv-power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        component_graph = connection_manager.get().component_graph
        component_ids = self._config.component_ids
        if component_ids:
            pv_components = component_graph.components(set(component_ids))
        else:
            pv_components = component_graph.dfs(
                self._get_grid_component(),
                set(),
                component_graph.is_pv_chain,
            )

        if not pv_components:
            _logger.warning(
                "Unable to find any PV components in the component graph. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no PV components, we have to send 0 values at the same
            # frequency as the other streams.  So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID,
                nones_are_zeros=True,
            )
            return builder.build()

        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(pv_components)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=(
                        primary_component.category != ComponentCategory.METER
                    ),
                    fallback=fallback_formula,
                )
        else:
            for idx, component in enumerate(pv_components):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    component.component_id,
                    nones_are_zeros=component.category != ComponentCategory.METER,
                )

        return builder.build()

    def _get_fallback_formulas(
        self, components: set[Component]
    ) -> dict[Component, FallbackFormulaMetricFetcher[Power] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the PV power.
        If it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the PV power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher`.

        Args:
            components: The PV components.

        Returns:
            A dictionary mapping primary components to their corresponding
                FallbackFormulaMetricFetcher.
        """
        fallbacks = self._get_metric_fallback_components(components)

        fallback_formulas: dict[
            Component, FallbackFormulaMetricFetcher[Power] | None
        ] = {}
        for primary_component, fallback_components in fallbacks.items():
            if len(fallback_components) == 0:
                fallback_formulas[primary_component] = None
                continue
            fallback_ids = [c.component_id for c in fallback_components]

            generator = PVPowerFormula(
                f"{self._namespace}_fallback_{fallback_ids}",
                self._channel_registry,
                self._resampler_subscription_sender,
                FormulaGeneratorConfig(
                    component_ids=set(fallback_ids),
                    allow_fallback=False,
                ),
            )

            fallback_formulas[primary_component] = FallbackFormulaMetricFetcher(
                generator
            )

        return fallback_formulas



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""The DataSourcing Actor."""

from frequenz.channels import Receiver

from ..._internal._channels import ChannelRegistry
from ...actor import Actor
from ._component_metric_request import ComponentMetricRequest
from .microgrid_api_source import MicrogridApiSource


class DataSourcingActor(Actor):
    """An actor that provides data streams of metrics as time series."""

    def __init__(
        self,
        request_receiver: Receiver[ComponentMetricRequest],
        registry: ChannelRegistry,
        *,
        name: str | None = None,
    ) -> None:
        """Create a `DataSourcingActor` instance.

        Args:
            request_receiver: A channel receiver to accept metric requests from.
            registry: A channel registry.  To be replaced by a singleton
                instance.
            name: The name of the actor. If `None`, `str(id(self))` will be used. This
                is used mostly for debugging purposes.
        """
        super().__init__(name=name)
        self._request_receiver = request_receiver
        self._microgrid_api_source = MicrogridApiSource(registry)

    async def _run(self) -> None:
        """Run the actor."""
        async for request in self._request_receiver:
            await self._microgrid_api_source.add_metric(request)



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Script with an example how to use BatteryPool."""


import asyncio
import logging
from datetime import timedelta

from frequenz.channels import merge

from frequenz.sdk import microgrid
from frequenz.sdk.actor import ResamplerConfig

MICROGRID_API_URL = "grpc://microgrid.sandbox.api.frequenz.io:62060"


async def main() -> None:
    """Create the battery pool, activate all formulas and listen for any update."""
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s %(name)s %(levelname)s:%(message)s"
    )
    # Reduce undesired noise from hpack
    logging.getLogger("hpack.hpack").setLevel(logging.INFO)

    await microgrid.initialize(
        MICROGRID_API_URL,
        resampler_config=ResamplerConfig(resampling_period=timedelta(seconds=1.0)),
    )

    battery_pool = microgrid.new_battery_pool(priority=5)
    receivers = [
        battery_pool.soc.new_receiver(limit=1),
        battery_pool.capacity.new_receiver(limit=1),
        # pylint: disable=protected-access
        battery_pool._system_power_bounds.new_receiver(limit=1),
        # pylint: enable=protected-access
    ]

    async for metric in merge(*receivers):
        print(f"Received new metric: {metric}")


asyncio.run(main())



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Ringbuffer dumping & loading functions."""

# For use of the class type hint inside the class itself.


import pickle
from os.path import exists

from .buffer import FloatArray, OrderedRingBuffer

FILE_FORMAT_VERSION: int = 1
"""Version of the latest file dumping/loading format."""


def load(path: str) -> OrderedRingBuffer[FloatArray] | None:
    """Load a ringbuffer from disk.

    Args:
        path: Path to the file where the data is stored.

    Raises:
        RuntimeError: when the file format version is unknown.

    Returns:
        `None` when the file doesn't exist, otherwise an instance of the
        `OrderedRingBuffer` class, loaded from disk.
    """
    if not exists(path):
        return None

    with open(path, mode="rb") as fileobj:
        instance: OrderedRingBuffer[FloatArray]
        file_format_version: int

        file_format_version, instance = pickle.load(fileobj)

    if file_format_version != FILE_FORMAT_VERSION:
        raise RuntimeError(
            f"Unknown file format version: {file_format_version}. Can load: {FILE_FORMAT_VERSION}"
        )

    return instance


def dump(  # noqa: DOC502 (OSError is raised indirectly by open and pickle.dump)
    ringbuffer: OrderedRingBuffer[FloatArray],
    path: str,
    file_format_version: int = FILE_FORMAT_VERSION,
) -> None:
    """Dump a ringbuffer to disk.

    Args:
        ringbuffer: Instance of the ringbuffer to dump.
        path: Path to where the data should be saved to.
        file_format_version: Version of the file format, optional.

    Raises:
        OSError: When the file cannot be opened or written.
    """
    with open(path, mode="wb+") as fileobj:
        pickle.dump((file_format_version, ringbuffer), fileobj)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Results from PowerDistributingActor."""


import dataclasses
from collections import abc

from frequenz.quantities import Power

from .request import Request


@dataclasses.dataclass
class _BaseResultMixin:
    """Base mixin class for reporting power distribution results."""

    request: Request
    """The user's request to which this message responds."""


@dataclasses.dataclass
class _BaseSuccessMixin:
    """Result returned when setting the power succeed for all components."""

    succeeded_power: Power
    """The part of the requested power that was successfully set."""

    succeeded_components: abc.Set[int]
    """The subset of components for which power was set successfully."""

    excess_power: Power
    """The part of the requested power that could not be fulfilled.

    This happens when the requested power is outside the available power bounds.
    """


# We need to put the _BaseSuccessMixin before _BaseResultMixin in the
# inheritance list to make sure that the _BaseResultMixin attributes appear
# before the _BaseSuccessMixin, otherwise the request attribute will be last
# in the dataclass constructor because of how MRO works.


@dataclasses.dataclass
class Success(_BaseSuccessMixin, _BaseResultMixin):  # Order matters here. See above.
    """Result returned when setting the power was successful for all components."""


@dataclasses.dataclass
class PartialFailure(_BaseSuccessMixin, _BaseResultMixin):
    """Result returned when some of the components had an error setting the power."""

    failed_power: Power
    """The part of the requested power that failed to be set."""

    failed_components: abc.Set[int]
    """The subset of batteries for which the request failed."""


@dataclasses.dataclass
class Error(_BaseResultMixin):
    """Result returned when an error occurred and power was not set at all."""

    msg: str
    """The error message explaining why error happened."""


@dataclasses.dataclass
class PowerBounds:
    """Inclusion and exclusion power bounds for the requested components."""

    inclusion_lower: float
    """The lower value of the inclusion power bounds for the requested components."""

    exclusion_lower: float
    """The lower value of the exclusion power bounds for the requested components."""

    exclusion_upper: float
    """The upper value of the exclusion power bounds for the requested components."""

    inclusion_upper: float
    """The upper value of the inclusion power bounds for the requested components."""


@dataclasses.dataclass
class OutOfBounds(_BaseResultMixin):
    """Result returned when the power was not set because it was out of bounds.

    This result happens when the originating request was done with
    `adjust_power = False` and the requested power is not within the available bounds.
    """

    bounds: PowerBounds
    """The power bounds for the requested components.

    If the requested power negative, then this value is the lower bound.
    Otherwise it is upper bound.
    """


Result = Success | PartialFailure | Error | OutOfBounds
"""Power distribution result.

Example: Handling power distribution results

    ```python
    from typing import assert_never

    from frequenz.sdk.actor.power_distributing import (
        Error,
        OutOfBounds,
        PartialFailure,
        Result,
        Success,
    )
    from frequenz.sdk.actor.power_distributing.request import Request
    from frequenz.sdk.actor.power_distributing.result import PowerBounds
    from frequenz.quantities import Power

    def handle_power_request_result(result: Result) -> None:
        match result:
            case Success() as success:
                print(f"Power request was successful: {success}")
            case PartialFailure() as partial_failure:
                print(f"Power request was partially successful: {partial_failure}")
            case OutOfBounds() as out_of_bounds:
                print(f"Power request was out of bounds: {out_of_bounds}")
            case Error() as error:
                print(f"Power request failed: {error}")
            case _ as unreachable:
                assert_never(unreachable)

    request = Request(
        namespace="TestChannel",
        power=Power.from_watts(123.4),
        component_ids={8, 18},
    )

    results: list[Result] = [
        Success(
            request,
            succeeded_power=Power.from_watts(123.4),
            succeeded_components={8, 18},
            excess_power=Power.zero(),
        ),
        PartialFailure(
            request,
            succeeded_power=Power.from_watts(103.4),
            succeeded_components={8},
            excess_power=Power.zero(),
            failed_components={18},
            failed_power=Power.from_watts(20.0),
        ),
        OutOfBounds(request, bounds=PowerBounds(0, 0, 0, 800)),
        Error(request, msg="The components are not available"),
    ]

    for r in results:
        handle_power_request_result(r)
    ```
"""



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Internal math tools."""

import math


def is_close_to_zero(value: float, abs_tol: float = 1e-9) -> bool:
    """Check if a floating point value is close to zero.

    A value of 1e-9 is a commonly used absolute tolerance to balance precision
    and robustness for floating-point numbers comparisons close to zero. Note
    that this is also the default value for the relative tolerance.
    For more technical details, see https://peps.python.org/pep-0485/#behavior-near-zero

    Args:
        value: the floating point value to compare to.
        abs_tol: the minimum absolute tolerance. Defaults to 1e-9.

    Returns:
        whether the floating point value is close to zero.
    """
    zero: float = 0.0
    return math.isclose(a=value, b=zero, abs_tol=abs_tol)



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Utility types and functions for internal use."""

from ._singleton_meta import SingletonMeta

# Explicitly declare the public API.
__all__ = ["SingletonMeta"]



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""Formula generator from component graph for Producer Power."""

import logging
from typing import Callable

from frequenz.client.microgrid import Component, ComponentCategory, ComponentMetricId
from frequenz.quantities import Power

from ....microgrid import connection_manager
from .._formula_engine import FormulaEngine
from ._fallback_formula_metric_fetcher import FallbackFormulaMetricFetcher
from ._formula_generator import (
    NON_EXISTING_COMPONENT_ID,
    FormulaGenerator,
    FormulaGeneratorConfig,
)
from ._simple_formula import SimplePowerFormula

_logger = logging.getLogger(__name__)


class ProducerPowerFormula(FormulaGenerator[Power]):
    """Formula generator from component graph for calculating the Producer Power.

    The producer power is calculated by summing up the power of all power producers,
    which are CHP and PV.
    """

    def generate(  # noqa: DOC502
        # * ComponentNotFound is raised indirectly by _get_grid_component()
        # * RuntimeError is raised indirectly by connection_manager.get()
        self,
    ) -> FormulaEngine[Power]:
        """Generate formula for calculating producer power from the component graph.

        Returns:
            A formula engine that will calculate the producer power.

        Raises:
            ComponentNotFound: If the component graph does not contain a producer power
                component.
            RuntimeError: If the grid component has a single successor that is not a
                meter.
        """
        builder = self._get_builder(
            "producer_power", ComponentMetricId.ACTIVE_POWER, Power.from_watts
        )

        component_graph = connection_manager.get().component_graph
        # if in the future we support additional producers, we need to add them to the lambda
        producer_components = component_graph.dfs(
            self._get_grid_component(),
            set(),
            lambda component: component_graph.is_pv_chain(component)
            or component_graph.is_chp_chain(component),
        )

        if not producer_components:
            _logger.warning(
                "Unable to find any producer components in the component graph. "
                "Subscribing to the resampling actor with a non-existing "
                "component id, so that `0` values are sent from the formula."
            )
            # If there are no producer components, we have to send 0 values at the same
            # frequency as the other streams.  So we subscribe with a non-existing
            # component id, just to get a `None` message at the resampling interval.
            builder.push_component_metric(
                NON_EXISTING_COMPONENT_ID, nones_are_zeros=True
            )
            return builder.build()

        is_not_meter: Callable[[Component], bool] = (
            lambda component: component.category != ComponentCategory.METER
        )

        if self._config.allow_fallback:
            fallbacks = self._get_fallback_formulas(producer_components)

            for idx, (primary_component, fallback_formula) in enumerate(
                fallbacks.items()
            ):
                if idx > 0:
                    builder.push_oper("+")

                # should only be the case if the component is not a meter
                builder.push_component_metric(
                    primary_component.component_id,
                    nones_are_zeros=is_not_meter(primary_component),
                    fallback=fallback_formula,
                )
        else:
            for idx, component in enumerate(producer_components):
                if idx > 0:
                    builder.push_oper("+")

                builder.push_component_metric(
                    component.component_id,
                    nones_are_zeros=is_not_meter(component),
                )

        return builder.build()

    def _get_fallback_formulas(
        self, components: set[Component]
    ) -> dict[Component, FallbackFormulaMetricFetcher[Power] | None]:
        """Find primary and fallback components and create fallback formulas.

        The primary component is the one that will be used to calculate the producer power.
        However, if it is not available, the fallback formula will be used instead.
        Fallback formulas calculate the producer power using the fallback components.
        Fallback formulas are wrapped in `FallbackFormulaMetricFetcher`.

        Args:
            components: The producer components.

        Returns:
            A dictionary mapping primary components to their FallbackFormulaMetricFetcher.
        """
        fallbacks = self._get_metric_fallback_components(components)

        fallback_formulas: dict[
            Component, FallbackFormulaMetricFetcher[Power] | None
        ] = {}

        for primary_component, fallback_components in fallbacks.items():
            if len(fallback_components) == 0:
                fallback_formulas[primary_component] = None
                continue

            fallback_ids = [c.component_id for c in fallback_components]
            generator = SimplePowerFormula(
                f"{self._namespace}_fallback_{fallback_ids}",
                self._channel_registry,
                self._resampler_subscription_sender,
                FormulaGeneratorConfig(
                    component_ids=set(fallback_ids),
                    allow_fallback=False,
                ),
            )

            fallback_formulas[primary_component] = FallbackFormulaMetricFetcher(
                generator
            )

        return fallback_formulas



# License: MIT
# Copyright © 2022 Frequenz Energy-as-a-Service GmbH

"""The logical component for calculating high level consumer metrics for a microgrid."""

import uuid

from frequenz.channels import Sender
from frequenz.quantities import Power

from .._internal._channels import ChannelRegistry
from ..microgrid._data_sourcing import ComponentMetricRequest
from .formula_engine import FormulaEngine
from .formula_engine._formula_engine_pool import FormulaEnginePool
from .formula_engine._formula_generators import ConsumerPowerFormula


class Consumer:
    """Calculate high level consumer metrics in a microgrid.

    Under normal circumstances this is expected to correspond to the gross
    consumption of the site excluding active parts and battery.

    Consumer provides methods for fetching power values from different points
    in the microgrid. These methods return `FormulaReceiver` objects, which can
    be used like normal `Receiver`s, but can also be composed to form
    higher-order formula streams.

    !!! note
        `Consumer` instances are not meant to be created directly by users.
        Use the [`microgrid.consumer`][frequenz.sdk.microgrid.consumer] method
        for creating `Consumer` instances.

    Example:
        ```python
        from datetime import timedelta

        from frequenz.sdk import microgrid
        from frequenz.sdk.timeseries import ResamplerConfig

        await microgrid.initialize(
            "grpc://127.0.0.1:50051",
            ResamplerConfig(resampling_period=timedelta(seconds=1.0))
        )

        consumer = microgrid.consumer()

        # Get a receiver for a builtin formula
        consumer_power_recv = consumer.power.new_receiver()
        async for consumer_power_sample in consumer_power_recv:
            print(consumer_power_sample)
        ```
    """

    _formula_pool: FormulaEnginePool
    """The formula engine pool to generate consumer metrics."""

    def __init__(
        self,
        channel_registry: ChannelRegistry,
        resampler_subscription_sender: Sender[ComponentMetricRequest],
    ) -> None:
        """Initialize the consumer formula generator.

        Args:
            channel_registry: The channel registry to use for the consumer.
            resampler_subscription_sender: The sender to use for resampler subscriptions.
        """
        namespace = f"consumer-{uuid.uuid4()}"
        self._formula_pool = FormulaEnginePool(
            namespace,
            channel_registry,
            resampler_subscription_sender,
        )

    @property
    def power(self) -> FormulaEngine[Power]:
        """Fetch the consumer power for the microgrid.

        This formula produces values that are in the Passive Sign Convention (PSC).

        It will start the formula engine to calculate consumer power if it is
        not already running.

        A receiver from the formula engine can be created using the
        `new_receiver` method.

        Returns:
            A FormulaEngine that will calculate and stream consumer power.
        """
        engine = self._formula_pool.from_power_formula_generator(
            "consumer_power",
            ConsumerPowerFormula,
        )
        assert isinstance(engine, FormulaEngine)
        return engine

    async def stop(self) -> None:
        """Stop all formula engines."""
        await self._formula_pool.stop()



# License: MIT
# Copyright © 2023 Frequenz Energy-as-a-Service GmbH

"""Base classes for the power manager."""

from __future__ import annotations

import abc
import dataclasses
import enum
import typing

from frequenz.quantities import Power

from ... import timeseries
from . import _bounds

if typing.TYPE_CHECKING:
    from ...timeseries._base_types import SystemBounds


@dataclasses.dataclass(frozen=True, kw_only=True)
class ReportRequest:
    """A request to start a reporting stream."""

    source_id: str
    """The source ID of the actor sending the request."""

    component_ids: frozenset[int]
    """The component IDs to report on."""

    priority: int
    """The priority of the actor ."""

    set_operating_point: bool
    """Whether this proposal sets the operating point power or the normal power."""

    def get_channel_name(self) -> str:
        """Get the channel name for the report request.

        Returns:
            The channel name to use to identify the corresponding report channel
                from the channel registry.
        """
        return f"power_manager.report.{self.component_ids=}.{self.priority=}"


@dataclasses.dataclass(frozen=True, kw_only=True)
class _Report:
    """Current PowerManager report for a set of components."""

    target_power: Power | None
    """The currently set power for the components."""

    _inclusion_bounds: timeseries.Bounds[Power] | None
    """The available inclusion bounds for the components, for the actor's priority.

    These bounds are adjusted to any restrictions placed by actors with higher
    priorities.
    """

    _exclusion_bounds: timeseries.Bounds[Power] | None
    """The exclusion bounds for the components.

    The power manager doesn't manage exclusion bounds, so these are aggregations of
    values reported by the microgrid API.

    These bounds are adjusted to any restrictions placed by actors with higher
    priorities.
    """

    @property
    def bounds(self) -> timeseries.Bounds[Power] | None:
        """The bounds for the components.

        These bounds are adjusted to any restrictions placed by actors with higher
        priorities.

        There might be exclusion zones within these bounds. If necessary, the
        `adjust_to_bounds` method may be used to check if a desired power value fits the
        bounds, or to get the closest possible power values that do fit the bounds.
        """
        return self._inclusion_bounds

    def adjust_to_bounds(self, power: Power) -> tuple[Power | None, Power | None]:
        """Adjust a power value to the bounds.

        This method can be used to adjust a desired power value to the power bounds
        available to the actor.

        If the given power value falls within the usable bounds, it will be returned
        unchanged.

        If it falls outside the usable bounds, the closest possible value on the
        corresponding side will be returned.  For example, if the given power is lower
        than the lowest usable power, only the lowest usable power will be returned, and
        similarly for the highest usable power.

        If the given power falls within an exclusion zone that's contained within the
        usable bounds, the closest possible power values on both sides will be returned.

        !!! note
            It is completely optional to use this method to adjust power values before
            proposing them, because the PowerManager will do this automatically.  This
            method is provided for convenience, and for granular control when there are
            two possible power values, both of which fall within the available bounds.

        Args:
            power: The power value to adjust.

        Returns:
            A tuple of the closest power values to the desired power that fall within
                the available bounds for the actor.
        """
        if self._inclusion_bounds is None:
            return None, None

        return _bounds.clamp_to_bounds(
            power,
            self._inclusion_bounds.lower,
            self._inclusion_bounds.upper,
            self._exclusion_bounds,
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Proposal:
    """A proposal for a set of components to be charged or discharged."""

    source_id: str
    """The source ID of the actor sending the request."""

    preferred_power: Power | None
    """The preferred power to be distributed to the components.

    If `None`, the preferred power of higher priority actors will get precedence.
    """

    bounds: timeseries.Bounds[Power | None]
    """The power bounds for the proposal.

    These bounds will apply to actors with a lower priority, and can be overridden by
    bounds from actors with a higher priority.  If None, the power bounds will be set to
    the maximum power of the components in the pool.  This is currently an experimental
    feature.
    """

    component_ids: frozenset[int]
    """The component IDs to distribute the power to."""

    priority: int
    """The priority of the actor sending the proposal."""

    creation_time: float
    """The loop time when the proposal is created.

    This is used by the power manager to determine the age of the proposal.
    """

    set_operating_point: bool
    """Whether this proposal sets the operating point power or the normal power."""

    def __lt__(self, other: Proposal) -> bool:
        """Compare two proposals by their priority.

        When they have the same priority, compare them by their source ID.

        Args:
            other: The other proposal to compare to.

        Returns:
            Whether this proposal has a higher priority than the other proposal.
        """
        return (self.priority < other.priority) or (
            self.priority == other.priority and self.source_id < other.source_id
        )

    def __eq__(self, other: object) -> bool:
        """Check if two proposals are equal.

        Equality is determined by the priority and source ID of the proposals, so
        two proposals are equal if they have the same priority and source ID, even
        if they have different power values or creation times.

        This is so that there is only one active proposal for each actor in the bucket,
        and older proposals are replaced by newer ones.

        Args:
            other: The other proposal to compare to.

        Returns:
            Whether the two proposals are equal.
        """
        if not isinstance(other, Proposal):
            return NotImplemented

        return self.priority == other.priority and self.source_id == other.source_id

    def __hash__(self) -> int:
        """Get the hash of the proposal.

        Returns:
            The hash of the proposal.
        """
        return hash((self.priority, self.source_id))


class Algorithm(enum.Enum):
    """The available algorithms for the power manager."""

    MATRYOSHKA = "matryoshka"


class BaseAlgorithm(abc.ABC):
    """The base class for algorithms."""

    @abc.abstractmethod
    def calculate_target_power(
        self,
        component_ids: frozenset[int],
        proposal: Proposal | None,
        system_bounds: SystemBounds,
        must_return_power: bool = False,
    ) -> Power | None:
        """Calculate and return the target power for the given components.

        Args:
            component_ids: The component IDs to calculate the target power for.
            proposal: If given, the proposal to added to the bucket, before the target
                power is calculated.
            system_bounds: The system bounds for the components in the proposal.
            must_return_power: If `True`, the algorithm must return a target power,
                even if it hasn't changed since the last call.

        Returns:
            The new target power for the components, or `None` if the target power
                didn't change.
        """

    @abc.abstractmethod
    def get_target_power(
        self,
        component_ids: frozenset[int],
    ) -> Power | None:
        """Get the target power for the given components.

        Args:
            component_ids: The component IDs to get the target power for.

        Returns:
            The target power for the given components, or `None` if there is no target
                power.
        """

    # The arguments for this method are tightly coupled to the `Matryoshka` algorithm.
    # It can be loosened up when more algorithms are added.
    @abc.abstractmethod
    def get_status(
        self,
        component_ids: frozenset[int],
        priority: int,
        system_bounds: SystemBounds,
    ) -> _Report:
        """Get the bounds for a set of components, for the given priority.

        Args:
            component_ids: The IDs of the components to get the bounds for.
            priority: The priority of the actor for which the bounds are requested.
            system_bounds: The system bounds for the components.

        Returns:
            The bounds for the components.
        """

    @abc.abstractmethod
    def drop_old_proposals(self, loop_time: float) -> None:
        """Drop old proposals.

        This method is called periodically by the power manager.

        Args:
            loop_time: The current loop time.
        """
